from __future__ import annotations

from enum import IntEnum
from typing import List

from pykotor.common.geometry import Vector2
from pykotor.common.language import LocalizedString
from pykotor.common.misc import Game, Color, ResRef
from pykotor.resource.formats.gff import GFF, GFFContent, GFFStruct, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class ARE:
    """
    Stores static area data.

    Attributes:
        tag: "Tag" field.
        name: "Name" field.
        alpha_test: "AlphaTest" field.
        camera_style: "CameraStyle" field.
        default_envmap: "DefaultEnvMap" field.
        grass_texture: "Grass_TexName" field.
        grass_density: "Grass_Density" field.
        grass_size: "Grass_QuadSize" field.
        grass_prob_ll: "Grass_Prob_LL" field.
        grass_prob_lr: "Grass_Prob_LR" field.
        grass_prob_ul: "Grass_Prob_UL" field.
        grass_prob_ur: "Grass_Prob_UR" field.
        fog_enabled: "SunFogOn" field.
        fog_near: "SunFogNear" field.
        fog_far: "SunFogFar" field.
        shadows: "SunShadows" field.
        shadow_opacity: "ShadowOpacity" field.
        wind_power: "WindPower" field.
        unescapable: "Unescapable" field.
        disable_transit: "DisableTransit" field.
        stealth_xp: "StealthXPEnabled" field.
        stealth_xp_loss: "StealthXPLoss" field.
        stealth_xp_max: "StealthXPMax" field.
        on_enter: "OnEnter" field.
        on_exit: "OnExit" field.
        on_heartbeat: "OnHeartbeat" field.
        on_user_defined: "OnUserDefined" field.
        sun_ambient: "SunAmbientColor" field.
        sun_diffuse: "SunDiffuseColor" field.
        dynamic_light: "DynAmbientColor" field.
        fog_color: "SunFogColor" field.
        grass_ambient: "Grass_Ambient" field.
        grass_diffuse: "Grass_Diffuse" field.

        dirty_argb_1: "DirtyARGBOne" field. KotOR 2 Only.
        dirty_argb_2: "DirtyARGBTwo" field. KotOR 2 Only.
        dirty_argb_3: "DirtyARGBThree" field. KotOR 2 Only.
        grass_emissive: "Grass_Emissive" field. KotOR 2 Only.
        chance_rain: "ChanceRain" field. KotOR 2 Only.
        chance_snow: "ChanceSnow" field. KotOR 2 Only.
        chance_lightning: "ChanceLightning" field. KotOR 2 Only.
        dirty_size_1: "DirtySizeOne" field. KotOR 2 Only.
        dirty_formula_1: "DirtyFormulaOne" field. KotOR 2 Only.
        dirty_func_1: "DirtyFuncOne" field. KotOR 2 Only.
        dirty_size_2: "DirtySizeTwo" field. KotOR 2 Only.
        dirty_formula_2: "DirtyFormulaTwo" field. KotOR 2 Only.
        dirty_func_2: "DirtyFuncTwo" field. KotOR 2 Only.
        dirty_size_3: "DirtySizeThree" field. KotOR 2 Only.
        dirty_formula_3: "DirtyFormulaThre" field. KotOR 2 Only.
        dirty_func_3: "DirtyFuncThree" field. KotOR 2 Only.

        comment: "Comments" field. Used in toolset only.

        unused_id: "ID" field. Not used by the game engine.
        creator_id: "Creator_ID" field. Not used by the game engine.
        flags: "Flags" field. Not used by the game engine.
        mod_spot_check: "ModSpotCheck" field. Not used by the game engine.
        mod_listen_check: "ModListenCheck" field. Not used by the game engine.
        moon_ambient: "MoonAmbientColor" field. Not used by the game engine.
        moon_diffuse: "MoonDiffuseColor" field. Not used by the game engine.
        moon_fog: "MoonFogOn" field. Not used by the game engine.
        moon_fog_near: "MoonFogNear" field. Not used by the game engine.
        moon_fog_far: "MoonFogFar" field. Not used by the game engine.
        moon_fog_color: "MoonFogColor" field. Not used by the game engine.
        moon_shadows: "MoonShadows" field. Not used by the game engine.
        is_night: "IsNight" field. Not used by the game engine.
        lighting_scheme: "LightingScheme" field. Not used by the game engine.
        day_night: "DayNightCycle" field. Not used by the game engine.
        loadscreen_id: "LoadScreenID" field. Not used by the game engine.
        no_rest: "NoRest" field. Not used by the game engine.
        no_hang_back: "NoHangBack" field. Not used by the game engine.
        player_only: "PlayerOnly" field. Not used by the game engine.
        player_vs_player: "PlayerVsPlayer" field. Not used by the game engine.
    """

    BINARY_TYPE = ResourceType.ARE

    def __init__(
            self
    ):
        self.alpha_test: float = 0.0
        self.camera_style: int = 0

        self.chance_lightning: int = 0
        self.chance_snow: int = 0
        self.chance_rain: int = 0

        self.comment: str = ""

        self.default_envmap: ResRef = ResRef.from_blank()

        self.disable_transit: bool = False

        self.dynamic_light: Color = Color.BLACK
        self.sun_ambient: Color = Color.BLACK
        self.sun_diffuse: Color = Color.BLACK
        self.shadow_opacity: int = 0
        self.shadows: bool = False

        self.fog_color: Color = Color.BLACK
        self.fog_near: float = 0
        self.fog_far: float = 0
        self.fog_enabled: bool = False

        self.dirty_argb_1: Color = Color.BLACK
        self.dirty_func_1: int = 0
        self.dirty_size_1: int = 0
        self.dirty_formula_1: int = 0

        self.dirty_argb_2: Color = Color.BLACK
        self.dirty_func_2: int = 0
        self.dirty_size_2: int = 0
        self.dirty_formula_2: int = 0

        self.dirty_argb_3: Color = Color.BLACK
        self.dirty_func_3: int = 0
        self.dirty_size_3: int = 0
        self.dirty_formula_3: int = 0

        self.grass_ambient: Color = Color.BLACK
        self.grass_diffuse: Color = Color.BLACK
        self.grass_emissive: Color = Color.BLACK
        self.grass_density: float = 0.0
        self.grass_size: float = 0.0
        self.grass_prob_ll: float = 0.0
        self.grass_prob_lr: float = 0.0
        self.grass_prob_ul: float = 0.0
        self.grass_prob_ur: float = 0.0
        self.grass_texture: ResRef = ResRef.from_blank()

        self.wind_power: AREWindPower = AREWindPower.Still

        self.on_enter: ResRef = ResRef.from_blank()
        self.on_exit: ResRef = ResRef.from_blank()
        self.on_heartbeat: ResRef = ResRef.from_blank()
        self.on_user_defined: ResRef = ResRef.from_blank()

        self.stealth_xp: bool = False
        self.stealth_xp_loss: int = 0
        self.stealth_xp_max: int = 0

        self.name: LocalizedString = LocalizedString.from_invalid()
        self.tag: str = ""
        self.unescapable: bool = False

        self.map_point_1: Vector2 = Vector2.from_null()
        self.map_point_2: Vector2 = Vector2.from_null()
        self.world_point_1: Vector2 = Vector2.from_null()
        self.world_point_2: Vector2 = Vector2.from_null()
        self.map_res_x: int = 0
        self.map_zoom: int = 0
        self.north_axis: ARENorthAxis = ARENorthAxis.PositiveX

        self.rooms: List[ARERoom] = []

        self.version: int = 0

        # Deprecated:
        self.unused_id: int = 0
        self.creator_id: int = 0
        self.flags: int = 0
        self.mod_spot_check: int = 0
        self.mod_listen_check: int = 0
        self.moon_ambient: int = 0
        self.moon_diffuse: int = 0
        self.moon_fog: int = 0
        self.moon_fog_near: float = 0.0
        self.moon_fog_far: float = 0.0
        self.moon_fog_color: int = 0
        self.moon_shadows: int = 0
        self.is_night: int = 0
        self.lighting_scheme: int = 0
        self.day_night: int = 0
        self.loadscreen_id: int = 0
        self.no_rest: int = 0
        self.no_hang_back: int = 0
        self.player_only: int = 0
        self.player_vs_player: int = 0


class ARERoom:
    def __init__(
            self,
            name: str,
            weather: bool,
            env_audio: int,
            force_rating: int,
            ambient_scale: float
    ):
        self.name: str = name
        self.weather: bool = weather
        self.env_audio: int = env_audio
        self.force_rating: int = force_rating
        self.ambient_scale: float = ambient_scale


class AREWindPower(IntEnum):
    Still = 0
    Weak = 1
    Strong = 2


class ARENorthAxis(IntEnum):
    PositiveY = 0
    NegativeY = 1
    PositiveX = 2
    NegativeX = 3


def construct_are(
        gff: GFF
) -> ARE:
    are = ARE()

    root = gff.root
    are.north_axis = ARENorthAxis(root.acquire("Map", GFFStruct()).acquire("NorthAxis", 0))
    are.map_zoom = root.acquire("Map", GFFStruct()).acquire("MapZoom", 0)
    are.map_res_x = root.acquire("Map", GFFStruct()).acquire("MapResX", 0)
    are.map_point_1 = Vector2(root.acquire("Map", GFFStruct()).acquire("MapPt1X", 0.0),
                              root.acquire("Map", GFFStruct()).acquire("MapPt1Y", 0.0))
    are.map_point_2 = Vector2(root.acquire("Map", GFFStruct()).acquire("MapPt2X", 0.0),
                              root.acquire("Map", GFFStruct()).acquire("MapPt2Y", 0.0))
    are.world_point_1 = Vector2(root.acquire("Map", GFFStruct()).acquire("WorldPt1X", 0.0),
                                root.acquire("Map", GFFStruct()).acquire("WorldPt1Y", 0.0))
    are.world_point_2 = Vector2(root.acquire("Map", GFFStruct()).acquire("WorldPt2X", 0.0),
                                root.acquire("Map", GFFStruct()).acquire("WorldPt2Y", 0.0))
    are.version = root.acquire("Version", 0)
    are.tag = root.acquire("Tag", "")
    are.name = root.acquire("Name", LocalizedString.from_invalid())
    are.comment = root.acquire("Comments", "")
    are.alpha_test = root.acquire("AlphaTest", 0.0)
    are.camera_style = root.acquire("CameraStyle", 0)
    are.default_envmap = root.acquire("DefaultEnvMap", ResRef.from_blank())
    are.grass_texture = root.acquire("Grass_TexName", ResRef.from_blank())
    are.grass_density = root.acquire("Grass_Density", 0.0)
    are.grass_size = root.acquire("Grass_QuadSize", 0.0)
    are.grass_prob_ll = root.acquire("Grass_Prob_LL", 0.0)
    are.grass_prob_lr = root.acquire("Grass_Prob_LR", 0.0)
    are.grass_prob_ul = root.acquire("Grass_Prob_UL", 0.0)
    are.grass_prob_ur = root.acquire("Grass_Prob_UR", 0.0)
    are.fog_enabled = root.acquire("SunFogOn", 0)
    are.fog_near = root.acquire("SunFogNear", 0.0)
    are.fog_far = root.acquire("SunFogFar", 0.0)
    are.shadows = root.acquire("SunShadows", 0)
    are.shadow_opacity = root.acquire("ShadowOpacity", 0)
    are.wind_power = root.acquire("WindPower", 0)
    are.unescapable = root.acquire("Unescapable", 0)
    are.disable_transit = root.acquire("DisableTransit", 0)
    are.stealth_xp = root.acquire("StealthXPEnabled", 0)
    are.stealth_xp_loss = root.acquire("StealthXPLoss", 0)
    are.stealth_xp_max = root.acquire("StealthXPMax", 0)
    are.on_enter = root.acquire("OnEnter", ResRef.from_blank())
    are.on_exit = root.acquire("OnExit", ResRef.from_blank())
    are.on_heartbeat = root.acquire("OnHeartbeat", ResRef.from_blank())
    are.on_user_defined = root.acquire("OnUserDefined", ResRef.from_blank())
    are.chance_rain = root.acquire("ChanceRain", 0)
    are.chance_snow = root.acquire("ChanceSnow", 0)
    are.chance_lightning = root.acquire("ChanceLightning", 0)
    are.dirty_size_1 = root.acquire("DirtySizeOne", 0)
    are.dirty_formula_1 = root.acquire("DirtyFormulaOne", 0)
    are.dirty_func_1 = root.acquire("DirtyFuncOne", 0)
    are.dirty_size_2 = root.acquire("DirtySizeTwo", 0)
    are.dirty_formula_2 = root.acquire("DirtyFormulaTwo", 0)
    are.dirty_func_2 = root.acquire("DirtyFuncTwo", 0)
    are.dirty_size_3 = root.acquire("DirtySizeThree", 0)
    are.dirty_formula_3 = root.acquire("DirtyFormulaThre", 0)
    are.dirty_func_3 = root.acquire("DirtyFuncThree", 0)
    are.unused_id = root.acquire("ID", 0)
    are.creator_id = root.acquire("Creator_ID", 0)
    are.flags = root.acquire("Flags", 0)
    are.mod_spot_check = root.acquire("ModSpotCheck", 0)
    are.mod_listen_check = root.acquire("ModListenCheck", 0)
    are.moon_ambient = root.acquire("MoonAmbientColor", 0)
    are.moon_diffuse = root.acquire("MoonDiffuseColor", 0)
    are.moon_fog = root.acquire("MoonFogOn", 0)
    are.moon_fog_near = root.acquire("MoonFogNear", 0.0)
    are.moon_fog_far = root.acquire("MoonFogFar", 0.0)
    are.moon_fog_color = root.acquire("MoonFogColor", 0)
    are.moon_shadows = root.acquire("MoonShadows", 0)
    are.is_night = root.acquire("IsNight", 0)
    are.lighting_scheme = root.acquire("LightingScheme", 0)
    are.day_night = root.acquire("DayNightCycle", 0)
    are.loadscreen_id = root.acquire("LoadScreenID", 0)
    are.no_rest = root.acquire("NoRest", 0)
    are.no_hang_back = root.acquire("NoHangBack", 0)
    are.player_only = root.acquire("PlayerOnly", 0)
    are.player_vs_player = root.acquire("PlayerVsPlayer", 0)

    are.sun_ambient = Color.from_rgb_integer(root.acquire("SunAmbientColor", 0))
    are.sun_diffuse = Color.from_rgb_integer(root.acquire("SunDiffuseColor", 0))
    are.dynamic_light = Color.from_rgb_integer(root.acquire("DynAmbientColor", 0))
    are.fog_color = Color.from_rgb_integer(root.acquire("SunFogColor", 0))
    are.grass_ambient = Color.from_rgb_integer(root.acquire("Grass_Ambient", 0))
    are.grass_diffuse = Color.from_rgb_integer(root.acquire("Grass_Diffuse", 0))

    are.grass_emissive = Color.from_rgb_integer(root.acquire("Grass_Emissive", 0))
    are.dirty_argb_1 = Color.from_rgb_integer(root.acquire("DirtyARGBOne", 0))
    are.dirty_argb_2 = Color.from_rgb_integer(root.acquire("DirtyARGBTwo", 0))
    are.dirty_argb_3 = Color.from_rgb_integer(root.acquire("DirtyARGBThree", 0))

    return are


def dismantle_are(
        are: ARE,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.ARE)

    root = gff.root

    map_struct = root.set_struct("Map", GFFStruct())
    map_struct.set_int32("MapZoom", are.map_zoom)
    map_struct.set_int32("MapResX", are.map_res_x)
    map_struct.set_int32("NorthAxis", are.north_axis.value)
    map_struct.set_single("MapPt1X", are.map_point_1.x)
    map_struct.set_single("MapPt1Y", are.map_point_1.y)
    map_struct.set_single("MapPt2X", are.map_point_2.x)
    map_struct.set_single("MapPt2Y", are.map_point_2.y)
    map_struct.set_single("WorldPt1X", are.world_point_1.x)
    map_struct.set_single("WorldPt1Y", are.world_point_1.y)
    map_struct.set_single("WorldPt2X", are.world_point_2.x)
    map_struct.set_single("WorldPt2Y", are.world_point_2.y)

    root.set_uint32("Version", are.version)

    root.set_uint32("SunAmbientColor", are.sun_ambient.rgb_integer())
    root.set_uint32("SunDiffuseColor", are.sun_diffuse.rgb_integer())
    root.set_uint32("DynAmbientColor", are.dynamic_light.rgb_integer())
    root.set_uint32("SunFogColor", are.fog_color.rgb_integer())
    root.set_uint32("Grass_Ambient", are.grass_ambient.rgb_integer())
    root.set_uint32("Grass_Diffuse", are.grass_diffuse.rgb_integer())

    root.set_string("Tag", are.tag)
    root.set_locstring("Name", are.name)
    root.set_string("Comments", are.comment)
    root.set_single("AlphaTest", are.alpha_test)
    root.set_int32("CameraStyle", are.camera_style)
    root.set_resref("DefaultEnvMap", are.default_envmap)
    root.set_resref("Grass_TexName", are.grass_texture)
    root.set_single("Grass_Density", are.grass_density)
    root.set_single("Grass_QuadSize", are.grass_size)
    root.set_single("Grass_Prob_LL", are.grass_prob_ll)
    root.set_single("Grass_Prob_LR", are.grass_prob_lr)
    root.set_single("Grass_Prob_UL", are.grass_prob_ul)
    root.set_single("Grass_Prob_UR", are.grass_prob_ur)
    root.set_uint8("SunFogOn", are.fog_enabled)
    root.set_single("SunFogNear", are.fog_near)
    root.set_single("SunFogFar", are.fog_far)
    root.set_uint8("SunShadows", are.shadows)
    root.set_uint8("ShadowOpacity", are.shadow_opacity)
    root.set_int32("WindPower", are.wind_power)
    root.set_uint8("Unescapable", are.unescapable)
    root.set_uint8("DisableTransit", are.disable_transit)
    root.set_uint8("StealthXPEnabled", are.stealth_xp)
    root.set_uint32("StealthXPLoss", are.stealth_xp_loss)
    root.set_uint32("StealthXPMax", are.stealth_xp_max)
    root.set_resref("OnEnter", are.on_enter)
    root.set_resref("OnExit", are.on_exit)
    root.set_resref("OnHeartbeat", are.on_heartbeat)
    root.set_resref("OnUserDefined", are.on_user_defined)

    if game == Game.K2:
        root.set_int32("DirtyARGBOne", are.dirty_argb_1.rgb_integer())
        root.set_int32("DirtyARGBTwo", are.dirty_argb_2.rgb_integer())
        root.set_int32("DirtyARGBThree", are.dirty_argb_3.rgb_integer())
        root.set_uint32("Grass_Emissive", are.grass_emissive.rgb_integer())

        root.set_int32("ChanceRain", are.chance_rain)
        root.set_int32("ChanceSnow", are.chance_snow)
        root.set_int32("ChanceLightning", are.chance_lightning)
        root.set_int32("DirtySizeOne", are.dirty_size_1)
        root.set_int32("DirtyFormulaOne", are.dirty_formula_1)
        root.set_int32("DirtyFuncOne", are.dirty_func_1)
        root.set_int32("DirtySizeTwo", are.dirty_size_2)
        root.set_int32("DirtyFormulaTwo", are.dirty_formula_2)
        root.set_int32("DirtyFuncTwo", are.dirty_func_2)
        root.set_int32("DirtySizeThree", are.dirty_size_3)
        root.set_int32("DirtyFormulaThre", are.dirty_formula_3)
        root.set_int32("DirtyFuncThree", are.dirty_func_3)

    if use_deprecated:
        root.set_int32("ID", are.unused_id)
        root.set_int32("Creator_ID", are.creator_id)
        root.set_uint32("Flags", are.flags)
        root.set_int32("ModSpotCheck", are.mod_spot_check)
        root.set_int32("ModListenCheck", are.mod_listen_check)
        root.set_uint32("MoonAmbientColor", are.moon_ambient)
        root.set_uint32("MoonDiffuseColor", are.moon_diffuse)
        root.set_uint8("MoonFogOn", are.moon_fog)
        root.set_single("MoonFogNear", are.moon_fog_near)
        root.set_single("MoonFogFar", are.moon_fog_far)
        root.set_uint32("MoonFogColor", are.moon_fog_color)
        root.set_uint8("MoonShadows", are.moon_shadows)
        root.set_uint8("IsNight", are.is_night)
        root.set_uint8("LightingScheme", are.lighting_scheme)
        root.set_uint8("DayNightCycle", are.day_night)
        root.set_uint16("LoadScreenID", are.loadscreen_id)
        root.set_uint8("NoRest", are.no_rest)
        root.set_uint8("NoHangBack", are.no_hang_back)
        root.set_uint8("PlayerOnly", are.player_only)
        root.set_uint8("PlayerVsPlayer", are.player_vs_player)

    return gff


def read_are(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> ARE:
    gff = read_gff(source, offset, size)
    are = construct_are(gff)
    return are


def write_are(
        are: ARE,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_are(are, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_are(
        are: ARE,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_are(are, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.ssf import SSF, SSFBinaryReader, SSFXMLReader, SSFBinaryWriter
from pykotor.resource.formats.ssf.io_ssf_xml import SSFXMLWriter
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceType


def detect_ssf(
        source: SOURCE_TYPES,
        offset: int = 0
) -> ResourceType:
    """
    Returns what format the SSF data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the SSF data.
        offset: Offset into the data.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.

    Returns:
        The format of the SSF data.
    """
    try:
        if isinstance(source, str):
            with BinaryReader.from_file(source, offset) as reader:
                file_format = ResourceType.SSF if reader.read_string(4) == "SSF " else ResourceType.SSF_XML
        elif isinstance(source, bytes) or isinstance(source, bytearray):
            file_format = ResourceType.SSF if source[:4].decode('ascii', 'ignore') == "SSF " else ResourceType.SSF_XML
        elif isinstance(source, BinaryReader):
            file_format = ResourceType.SSF if source.read_string(4) == "SSF " else ResourceType.SSF_XML
            source.skip(-4)
        else:
            file_format = ResourceType.INVALID
    except (FileNotFoundError, PermissionError, IsADirectoryError) as e:
        raise e
    except IOError:
        file_format = ResourceType.INVALID

    return file_format


def read_ssf(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> SSF:
    """
    Returns an SSF instance from the source. The file format (SSF or SSF_XML) is automatically determined before parsing
    the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or the format could not be determined.

    Returns:
        An SSF instance.
    """
    file_format = detect_ssf(source, offset)

    if file_format is ResourceType.INVALID:
        raise ValueError("Failed to determine the format of the GFF file.")

    if file_format == ResourceType.SSF:
        return SSFBinaryReader(source, offset, size).load()
    elif file_format == ResourceType.SSF_XML:
        return SSFXMLReader(source, offset, size).load()


def write_ssf(
        ssf: SSF,
        target: TARGET_TYPES,
        file_format: ResourceType = ResourceType.SSF
) -> None:
    """
    Writes the SSF data to the target location with the specified format (SSF or SSF_XML).

    Args:
        ssf: The SSF file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.SSF:
        SSFBinaryWriter(ssf, target).write()
    elif file_format == ResourceType.SSF_XML:
        SSFXMLWriter(ssf, target).write()
    else:
        raise ValueError("Unsupported format specified; use SSF or SSF_XML.")


def bytes_ssf(
        ssf: SSF,
        file_format: ResourceType = ResourceType.SSF
) -> bytes:
    """
    Returns the SSF data in the specified format (SSF or SSF_XML) as a bytes object.

    This is a convience method that wraps the write_ssf() method.

    Args:
        ssf: The target SSF object.
        file_format: The file format.

    Raises:
        ValueError: If the specified format was unsupported.

    Returns:
        The SSF data.
    """
    data = bytearray()
    write_ssf(ssf, data, file_format)
    return data



from pykotor.resource.formats.bwm.bwm_data import BWM, BWMType, BWMFace
from pykotor.resource.formats.bwm.io_bwm import BWMBinaryReader, BWMBinaryWriter
from pykotor.resource.formats.bwm.bwm_auto import write_bwm, read_bwm



from __future__ import annotations

import io
from typing import Optional
from xml.etree import ElementTree

from pykotor.common.language import Language
from pykotor.common.misc import ResRef
from pykotor.resource.formats.tlk.tlk_data import TLK
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter, autoclose


class TLKXMLReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = None
    ):
        super().__init__(source, offset, size)
        self._tlk: Optional[TLK] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> TLK:
        self._tlk = TLK()

        data = self._reader.read_bytes(self._reader.size()).decode()
        xml = ElementTree.parse(io.StringIO(data)).getroot()

        self._tlk.language = Language(int(xml.get("language")))
        self._tlk.resize(len(xml))
        for string in xml:
            index = int(string.get("id"))
            self._tlk.entries[index].text = string.text
            self._tlk.entries[index].voiceover = ResRef(string.get("sound")) if string.get("sound") else ResRef.from_blank()

        return self._tlk


class TLKXMLWriter(ResourceWriter):
    def __init__(
            self,
            tlk: TLK,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._xml: ElementTree.Element = ElementTree.Element("xml")
        self._tlk: TLK = tlk

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        self._xml.tag = "tlk"
        self._xml.set("language", str(self._tlk.language.value))

        for stringref, entry in self._tlk:
            element = ElementTree.Element("string")
            element.text = entry.text
            element.set("id", str(stringref))
            if entry.voiceover:
                element.set("sound", entry.voiceover.get())
            self._xml.append(element)

        ElementTree.indent(self._xml)
        self._writer.write_bytes(ElementTree.tostring(self._xml))



from __future__ import annotations

import glm
from OpenGL.GL import shaders, glGenTextures, glTexImage2D, glGetUniformLocation, glUniformMatrix4fv, glUniform4fv, \
    glUniform3fv
from OpenGL.GL.framebufferobjects import glGenerateMipmap
from OpenGL.GL.shaders import GL_FALSE
from OpenGL.raw.GL.EXT.texture_compression_s3tc import GL_COMPRESSED_RGB_S3TC_DXT1_EXT, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT
from OpenGL.raw.GL.VERSION.GL_1_0 import GL_TEXTURE_2D, glTexParameteri, GL_RGB, GL_UNSIGNED_BYTE, \
    GL_CLAMP, GL_LINEAR, GL_TEXTURE_WRAP_S, GL_TEXTURE_WRAP_T, GL_TEXTURE_MIN_FILTER, GL_TEXTURE_MAG_FILTER, GL_REPEAT, \
    GL_RGBA, GL_NEAREST_MIPMAP_LINEAR
from OpenGL.raw.GL.VERSION.GL_1_1 import glBindTexture
from OpenGL.raw.GL.VERSION.GL_1_3 import glCompressedTexImage2D
from OpenGL.raw.GL.VERSION.GL_2_0 import GL_VERTEX_SHADER, GL_FRAGMENT_SHADER, glUseProgram, glUniform1i
from glm import mat4, vec4, vec3
from pykotor.resource.formats.tpc import TPC, TPCTextureFormat

KOTOR_VSHADER = """
#version 330 core

layout (location = 0) in vec3 flags;
layout (location = 1) in vec3 position;
layout (location = 2) in vec3 normal;
layout (location = 3) in vec3 uv;
layout (location = 4) in vec3 uv2;

out vec2 diffuse_uv;
out vec2 lightmap_uv;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

void main()
{
    gl_Position = projection * view * model *  vec4(position, 1.0);
    diffuse_uv = vec2(uv.x, uv.y);
    lightmap_uv = vec2(uv2.x, uv2.y);
}
"""


KOTOR_FSHADER = """
#version 420
in vec2 diffuse_uv;
in vec2 lightmap_uv;

out vec4 FragColor;

layout(binding = 0) uniform sampler2D diffuse;
layout(binding = 1) uniform sampler2D lightmap;
uniform int enableLightmap;

void main()
{
    vec4 diffuseColor = texture(diffuse, diffuse_uv);
    vec4 lightmapColor = texture(lightmap, lightmap_uv);
    
    if (enableLightmap == 1) {
        FragColor = mix(diffuseColor, lightmapColor, 0.5);
    } else {
        FragColor = diffuseColor;
    }
}
"""


PICKER_VSHADER = """
#version 330 core

layout (location = 1) in vec3 position;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

void main()
{
    gl_Position = projection * view * model *  vec4(position, 1.0);
}
"""


PICKER_FSHADER = """
#version 330

uniform vec3 colorId;

out vec4 FragColor;

void main()
{
    FragColor = vec4(colorId, 1.0);
}
"""


PLAIN_VSHADER = """
#version 330 core

layout (location = 1) in vec3 position;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

void main()
{
    gl_Position = projection * view * model *  vec4(position, 1.0);
}
"""


PLAIN_FSHADER = """
#version 330

uniform vec4 color;

out vec4 FragColor;

void main()
{
    FragColor = color;
}
"""


class Shader:
    def __init__(self, vshader: str, fshader: str):
        vertex_shader = shaders.compileShader(vshader, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(fshader, GL_FRAGMENT_SHADER)
        self._id: int = shaders.compileProgram(vertex_shader, fragment_shader)

    def use(self) -> None:
        glUseProgram(self._id)

    def uniform(self, uniform_name: str) -> None:
        return glGetUniformLocation(self._id, uniform_name)

    def set_matrix4(self, uniform: str, matrix: mat4):
        glUniformMatrix4fv(self.uniform(uniform), 1, GL_FALSE, glm.value_ptr(matrix))

    def set_vector4(self, uniform: str, vector: vec4):
        glUniform4fv(self.uniform(uniform), 1, glm.value_ptr(vector))

    def set_vector3(self, uniform: str, vector: vec3):
        glUniform3fv(self.uniform(uniform), 1, glm.value_ptr(vector))

    def set_bool(self, uniform: str, boolean: bool):
        glUniform1i(self.uniform(uniform), boolean)


class Texture:
    def __init__(self, tex_id: int):
        self._id = tex_id

    @classmethod
    def from_tpc(cls, tpc: TPC) -> Texture:
        width, height, tpc_format, data = tpc.get(0)
        imageSize = len(data)

        gl_id = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, gl_id)

        if tpc_format == TPCTextureFormat.DXT1:
            glCompressedTexImage2D(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGB_S3TC_DXT1_EXT, width, height, 0, imageSize, data)
        if tpc_format == TPCTextureFormat.DXT5:
            glCompressedTexImage2D(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, width, height, 0, imageSize, data)
        if tpc_format == TPCTextureFormat.RGB:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, data)
        if tpc_format == TPCTextureFormat.RGBA:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data)

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glGenerateMipmap(GL_TEXTURE_2D)

        return Texture(gl_id)

    @classmethod
    def from_color(cls, r: int = 0, g: int = 0, b: int = 0) -> Texture:
        gl_id = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, gl_id)

        pixels = []
        for i in range(64):
            for j in range(64):
                pixels.extend([r, g, b])

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 64, 64, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        return Texture(gl_id)

    def use(self) -> None:
        glBindTexture(GL_TEXTURE_2D, self._id)



from pykotor.resource.formats.ncs.ncs_data import NCS, NCSInstruction, NCSInstructionType, NCSInstructionQualifier, \
    NCSInstructionTypeValue, NCSByteCode
from pykotor.resource.formats.ncs.io_ncs import NCSBinaryReader, NCSBinaryWriter
from pykotor.resource.formats.ncs.ncs_auto import read_ncs, write_ncs



"""
This module holds classes relating to geometry.
"""
from __future__ import annotations

import math
from enum import IntEnum
from typing import List


class Vector2:
    """
    Represents a 2 dimensional vector.

    Attributes:
        x: The x component.
        y: The y component.
    """

    def __init__(
            self,
            x: float,
            y: float
    ):
        self.x: float = x
        self.y: float = y

    def __iter__(
            self
    ):
        return iter((self.x, self.y))

    def __repr__(
            self
    ):
        return "Vector2({}, {})".format(self.x, self.y)

    def __str__(
            self
    ):
        """
        Returns the individual components separated by whitespace.
        """
        return "{} {}".format(self.x, self.y)

    def __eq__(
            self,
            other
    ):
        """
        Two Vector2 components are equal if their components are approximately the same.
        """
        if not isinstance(other, Vector2):
            return NotImplemented

        isclose_x = math.isclose(self.x, other.x)
        isclose_y = math.isclose(self.y, other.y)
        return isclose_x and isclose_y

    def __add__(
            self,
            other
    ):
        """
        Adds the components of two Vector2 objects.
        """
        if not isinstance(other, Vector2):
            return NotImplemented

        new = Vector2.from_vector2(self)
        new.x += other.x
        new.y += other.y
        return new

    def __sub__(
            self,
            other
    ):
        """
        Subtracts the components of two Vector2 objects.
        """
        if not isinstance(other, Vector2):
            return NotImplemented

        new = Vector2.from_vector2(self)
        new.x -= other.x
        new.y -= other.y
        return new

    def __mul__(
            self,
            other
    ):
        """
        Multiplies the components by a scalar integer.
        """
        if not isinstance(other, int):
            return NotImplemented

        new = Vector2.from_vector2(self)
        new.x *= other
        new.y *= other
        return new

    def __truediv__(
            self,
            other
    ):
        if isinstance(other, int):
            new = Vector2.from_vector2(self)
            new.x /= other
            new.y /= other
            return new
        else:
            return NotImplemented

    def __getitem__(
            self,
            item
    ):
        if isinstance(item, int):
            if item == 0:
                return self.x
            elif item == 1:
                return self.y
        else:
            return NotImplemented

    def __setitem__(
            self,
            key,
            value
    ):
        if isinstance(key, int) and (isinstance(value, float) or isinstance(value, int)):
            if key == 0:
                self.x = value
            elif key == 1:
                self.y = value
        else:
            return NotImplemented

    @classmethod
    def from_vector2(
            cls,
            other: Vector2
    ) -> Vector2:
        """
        Returns a duplicate of the specified vector.

        Args:
            other: The vector to be duplicated.

        Returns:
            A new Vector2 object.
        """
        return Vector2(other.x, other.y)

    @classmethod
    def from_vector3(
            cls,
            other: Vector3
    ) -> Vector2:
        """
        Returns a Vector2 object from the Vector3 object, discarding the Z-component.

        Args:
            other: The vector to be copied.

        Returns:
            A new Vector2 object.
        """
        return Vector2(other.x, other.y)

    @classmethod
    def from_vector4(
            cls,
            other: Vector4
    ) -> Vector2:
        """
        Returns a Vector2 object from the Vector4 object, discarding the Z-component and W-components.

        Args:
            other: The vector to be copied.

        Returns:
            A new Vector2 object.
        """
        return Vector2(other.x, other.y)

    @classmethod
    def from_null(
            cls
    ) -> Vector2:
        """
        Returns a new vector with a magnitude of zero.

        Returns:
            A new Vector2 instance.
        """
        return Vector2(0.0, 0.0)

    @classmethod
    def from_angle(
            cls,
            angle: float
    ) -> Vector2:
        """
        Returns a unit vector based on the specified angle.

        Args:
            angle: The angle of the new vector in radians.

        Returns:
            A new Vector2 instance.
        """
        x = math.cos(angle)
        y = math.sin(angle)
        return Vector2(x, y).normal()

    def set(
            self,
            x: float,
            y: float
    ) -> None:
        """
        Sets the components of the vector.

        Args:
            x: The new x component.
            y: The new y component.
        """
        self.x = x
        self.y = y

    def normalize(
            self
    ) -> None:
        """
        Normalizes the vector so that the magnitude is equal to one while maintaining the same angle.
        """
        magnitude = self.magnitude()
        if magnitude == 0:
            self.x = 0
            self.y = 0
        else:
            self.x /= magnitude
            self.y /= magnitude

    def magnitude(
            self
    ) -> float:
        """
        Returns the magnitude of the vector.

        Returns:
            The magnitude of the vector.
        """
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def normal(
            self
    ) -> Vector2:
        vec2 = Vector2.from_vector2(self)
        vec2.normalize()
        return vec2

    def dot(
            self,
            other: Vector2
    ) -> float:
        """
        Returns the dot product between the two specified vectors.

        Args:
            other: The other vector.

        Returns:
            The dot product.
        """
        a = self.x * other.x
        b = self.y * other.y
        return a + b

    def distance(
            self,
            other: Vector2
    ) -> float:
        """
        Returns the distance between two vectors.

        Args:
            other: The other vector.

        Returns:
            The distance between the vectors.
        """
        a = (self.x - other.x) ** 2
        b = (self.y - other.y) ** 2
        return math.sqrt(a + b)

    def within(
            self,
            container: List
    ) -> bool:
        """
        Checks to see if the same Vector2 object in located in the specified list. This differs from using the 'in'
        keyword as that will return True for Vector2 objects that have simular coordinates.

        Args:
            container: The list to search in.

        Returns:
            True if the Vector2 exists in the container.
        """
        for item in container:
            if item is self:
                return True
        return False

    def angle(
            self
    ) -> float:
        """
        Returns the angle of the vector.

        Returns:
            The angle of the vector in radians.
        """
        return math.atan2(self.y, self.x)


class Vector3:
    """
    Represents a 3 dimensional vector.

    Attributes:
        x: The x component.
        y: The y component.
        z: The z component.
    """

    def __init__(
            self,
            x: float,
            y: float,
            z: float
    ):
        self.x: float = x
        self.y: float = y
        self.z: float = z

    def __iter__(
            self
    ):
        return iter((self.x, self.y, self.z))

    def __repr__(
            self
    ):
        return "Vector3({}, {}, {})".format(self.x, self.y, self.z)

    def __str__(
            self
    ):
        """
        Returns the individual components as a string separated by whitespace.
        """
        return "{} {} {}".format(self.x, self.y, self.z)

    def __eq__(
            self,
            other
    ):
        """
        Two Vector3 components are equal if their components are approximately the same.
        """
        if not isinstance(other, Vector3):
            return NotImplemented

        isclose_x = math.isclose(self.x, other.x)
        isclose_y = math.isclose(self.y, other.y)
        isclose_z = math.isclose(self.z, other.z)
        return isclose_x and isclose_y and isclose_z

    def __add__(
            self,
            other
    ):
        """
        Adds the components of two Vector3 objects.
        """
        if not isinstance(other, Vector3):
            return NotImplemented

        new = Vector3.from_vector3(self)
        new.x += other.x
        new.y += other.y
        new.z += other.z
        return new

    def __sub__(
            self,
            other
    ):
        """
        Subtracts the components of two Vector3 objects.
        """
        if not isinstance(other, Vector3):
            return NotImplemented

        new = Vector3.from_vector3(self)
        new.x -= other.x
        new.y -= other.y
        new.z -= other.z
        return new

    def __mul__(
            self,
            other
    ):
        """
        Multiplies the components by a scalar integer.
        """
        if isinstance(other, int) or isinstance(other, float):
            new = Vector3.from_vector3(self)
            new.x *= other
            new.y *= other
            new.z *= other
            return new
        else:
            return NotImplemented

    def __truediv__(
            self,
            other
    ):
        if isinstance(other, int):
            new = Vector3.from_vector3(self)
            new.x /= other
            new.y /= other
            new.z /= other
            return new
        else:
            return NotImplemented

    def __getitem__(
            self,
            item
    ):
        if isinstance(item, int):
            if item == 0:
                return self.x
            elif item == 1:
                return self.y
            elif item == 2:
                return self.z
        else:
            return NotImplemented

    def __setitem__(
            self,
            key,
            value
    ):
        if isinstance(key, int) and (isinstance(value, float) or isinstance(value, int)):
            if key == 0:
                self.x = value
            elif key == 1:
                self.y = value
            elif key == 2:
                self.z = value
        else:
            return NotImplemented

    @classmethod
    def from_vector2(
            cls,
            other: Vector2
    ) -> Vector3:
        """
        Returns a Vector3 object from the Vector2 object, setting the Z-component to zero.

        Args:
            other: The vector to be copied.

        Returns:
            A new Vector3 object.
        """
        return Vector3(other.x, other.y, 0.0)

    @classmethod
    def from_vector3(
            cls,
            other: Vector3
    ) -> Vector3:
        """
        Returns a duplicate of the specified vector.

        Args:
            other: The vector to be duplicated.

        Returns:
            A new Vector3 instance.
        """
        return Vector3(other.x, other.y, other.z)

    @classmethod
    def from_vector4(
            cls,
            other: Vector4
    ) -> Vector3:
        """
        Returns a Vector3 object from the Vector4 object, discarding the W-component.

        Args:
            other: The vector to be copied.

        Returns:
            A new Vector3 object.
        """
        return Vector3(other.x, other.y, other.z)

    @classmethod
    def from_null(
            cls
    ) -> Vector3:
        """
        Returns a new vector with a magnitude of zero.

        Returns:
            A new Vector3 instance.
        """
        return Vector3(0.0, 0.0, 0.0)

    def set(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        """
        Sets the components of the vector.

        Args:
            x: The new x component.
            y: The new y component.
            z: The new y component.
        """
        self.x = x
        self.y = y
        self.z = z

    def normalize(
            self
    ) -> None:
        """
        Normalizes the vector so that the magnitude is equal to one while maintaining the same angle.
        """
        magnitude = self.magnitude()
        if magnitude == 0:
            self.x = 0
            self.y = 0
            self.z = 0
        else:
            self.x /= magnitude
            self.y /= magnitude
            self.z /= magnitude

    def magnitude(
            self
    ) -> float:
        """
        Returns the magnitude of the vector.

        Returns:
            The magnitude of the vector.
        """
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)

    def normal(
            self
    ) -> Vector3:
        vec3 = Vector3.from_vector3(self)
        vec3.normalize()
        return vec3

    def dot(
            self,
            other: Vector3
    ) -> float:
        """
        Returns the dot product between the two specified vectors.

        Args:
            other: The other vector.

        Returns:
            The dot product.
        """
        a = self.x * other.x
        b = self.y * other.y
        c = self.z * other.z
        return a + b + c

    def distance(
            self,
            other: Vector3
    ) -> float:
        """
        Returns the distance between two vectors.

        Args:
            other: The other vector.

        Returns:
            The distance between the vectors.
        """
        a = (self.x - other.x) ** 2
        b = (self.y - other.y) ** 2
        c = (self.z - other.z) ** 2
        return math.sqrt(a + b + c)

    def within(
            self,
            container: List
    ) -> bool:
        """
        Checks to see if the same Vector3 object in located in the specified list. This differs from using the 'in'
        keyword as that will return True for Vector3 objects that have simular coordinates.

        Args:
            container: The list to search in.

        Returns:
            True if the Vector3 exists in the container.
        """
        for item in container:
            if item is self:
                return True
        return False


class Vector4:
    """
    Represents a 4 dimensional vector.

    Attributes:
        x: The x component.
        y: The y component.
        z: The z component.
        w: The w component.
    """

    def __init__(
            self,
            x: float,
            y: float,
            z: float,
            w: float
    ):
        self.x: float = x
        self.y: float = y
        self.z: float = z
        self.w: float = w

    def __iter__(
            self
    ):
        return iter((self.x, self.y, self.z, self.w))

    def __repr__(
            self
    ):
        return "Vector4({}, {}, {}, {})".format(self.x, self.y, self.z, self.w)

    def __str__(
            self
    ):
        """
        Returns the individual components as a string separated by whitespace.
        """
        return "{} {} {} {}".format(self.x, self.y, self.z, self.w)

    def __eq__(
            self,
            other
    ):
        """
        Two Vector4 components are equal if their components are approximately the same.
        """
        if not isinstance(other, Vector4):
            return NotImplemented

        isclose_x = math.isclose(self.x, other.x)
        isclose_y = math.isclose(self.y, other.y)
        isclose_z = math.isclose(self.z, other.z)
        isclose_w = math.isclose(self.w, other.w)
        return isclose_x and isclose_y and isclose_z and isclose_w

    def __add__(
            self,
            other
    ):
        """
        Adds the components of two Vectorr objects.
        """
        if not isinstance(other, Vector4):
            return NotImplemented

        new = Vector4.from_vector4(self)
        new.x += other.x
        new.y += other.y
        new.z += other.z
        new.w += other.w
        return new

    def __sub__(
            self,
            other
    ):
        """
        Subtracts the components of two Vector4 objects.
        """
        if not isinstance(other, Vector4):
            return NotImplemented

        new = Vector4.from_vector4(self)
        new.x -= other.x
        new.y -= other.y
        new.z -= other.z
        new.w -= other.w
        return new

    def __mul__(
            self,
            other
    ):
        """
        Multiplies the components by a scalar integer.
        """
        if not isinstance(other, int):
            return NotImplemented

        new = Vector4.from_vector4(self)
        new.x *= other
        new.y *= other
        new.z *= other
        new.w *= other
        return new

    @classmethod
    def from_vector2(
            cls,
            other: Vector2
    ) -> Vector4:
        """
        Returns a Vector4 object from the Vector2 object, setting the Z-component and W-component to zero.

        Args:
            other: The vector to be copied.

        Returns:
            A new Vector4 object.
        """
        return Vector4(other.x, other.y, 0.0, 0.0)

    @classmethod
    def from_vector3(
            cls,
            other: Vector3
    ) -> Vector4:
        """
        Returns a Vector4 object from the Vector3 object, setting the W-component to zero.

        Args:
            other: The vector to be copied.

        Returns:
            A new Vector4 object.
        """
        return Vector4(other.x, other.y, other.z, 0.0)

    @classmethod
    def from_vector4(
            cls,
            other: Vector4
    ) -> Vector4:
        """
        Returns a duplicate of the specified vector.

        Args:
            other: The vector to be duplicated.

        Returns:
            A new Vector4 instance.
        """
        return Vector4(other.x, other.y, other.z, other.w)

    @classmethod
    def from_null(
            cls
    ) -> Vector4:
        """
        Returns a new vector with a magnitude of zero.

        Returns:
            A new Vector4 instance.
        """
        return Vector4(0.0, 0.0, 0.0, 0.0)

    @classmethod
    def from_compressed(
            cls,
            data: int
    ) -> Vector4:
        x = 1 - (data & 0x7FF) / 1023
        y = 1 - ((data >> 11) & 0x7FF) / 1023
        z = 1 - (data >> 22) / 511

        temp = x ** 2 + y ** 2 + z ** 2
        if temp < 1.0:
            w = -math.sqrt(1.0 - temp)
        else:
            temp = math.sqrt(temp)
            x /= temp
            y /= temp
            z /= temp
            w = 0

        return Vector4(x, y, z, w)

    @classmethod
    def from_euler(
            cls,
            x: float,
            y: float,
            z: float
    ):
        """
        Creates a Vector3 object from x/y/z rotations (in radians).

        Args:
            x: Roll
            y: Pitch
            z: Yaw

        Returns:
            A new Vector3 object.
        """
        roll = x
        pitch = y
        yaw = z

        qx = math.sin(roll/2) * math.cos(pitch/2) * math.cos(yaw/2) - math.cos(roll/2) * math.sin(pitch/2) * math.sin(yaw/2)
        qy = math.cos(roll/2) * math.sin(pitch/2) * math.cos(yaw/2) + math.sin(roll/2) * math.cos(pitch/2) * math.sin(yaw/2)
        qz = math.cos(roll/2) * math.cos(pitch/2) * math.sin(yaw/2) - math.sin(roll/2) * math.sin(pitch/2) * math.cos(yaw/2)
        qw = math.cos(roll/2) * math.cos(pitch/2) * math.cos(yaw/2) + math.sin(roll/2) * math.sin(pitch/2) * math.sin(yaw/2)

        return Vector4(qx, qy, qz, qw)

    def to_euler(
            self
    ) -> Vector3:
        t0 = 2.0 * (self.w*self.x + self.y*self.z)
        t1 = 1 - 2*(self.x*self.x + self.y*self.y)
        roll = math.atan2(t0, t1)

        t2 = 2 * (self.w*self.y - self.z*self.x)
        t2 = 1 if t2 > 1 else t2
        t2 = -1 if t2 < -1 else t2
        pitch = math.asin(t2)

        t3 = 2 * (self.w*self.z + self.x*self.y)
        t4 = 1 - 2*(self.y*self.y + self.z*self.z)
        yaw = math.atan2(t3, t4)

        return Vector3(roll, pitch, yaw)

    def magnitude(
            self
    ) -> float:
        """
        Returns the magnitude of the vector.

        Returns:
            The magnitude of the vector.
        """
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2 + self.w ** 2)

    def normalize(
            self
    ) -> Vector4:
        """
        Normalizes the vector so that the magnitude is equal to one while maintaining the same angle.

        Returns:
            The same vector.
        """
        magnitude = self.magnitude()
        if magnitude == 0:
            self.x = 0
            self.y = 0
            self.z = 0
            self.w = 0
        else:
            self.x /= magnitude
            self.y /= magnitude
            self.z /= magnitude
            self.w /= magnitude
        return self

    def set(
            self,
            x: float,
            y: float,
            z: float,
            w: float
    ) -> None:
        """
        Sets the components of the vector.

        Args:
            x: The new x component.
            y: The new y component.
            z: The new y component.
            w: The new w component.
        """
        self.x = x
        self.y = y
        self.z = z
        self.w = w


class AxisAngle:
    """
    Represents a rotation in 3D space.

    Attributes:
        axis: The axis of rotation.
        angle: The magnitude of the rotation.
    """

    def __init__(
            self,
            axis: Vector3,
            angle: float
    ):
        self.axis: Vector3 = axis
        self.angle: float = angle

    @classmethod
    def from_quaternion(
            cls,
            quaternion: Vector4
    ) -> AxisAngle:
        """
        Returns a AxisAngle converted from a Vector4 quaternion.

        Args:
            quaternion: The quaternion.

        Returns:
            A new AxisAngle instance.
        """
        aa = AxisAngle.from_null()
        quaternion.normalize()

        aa.angle = 2.0 * math.atan2(Vector3(quaternion.x, quaternion.y, quaternion.z).magnitude(), quaternion.w)

        s = math.sin(aa.angle / 2.0)
        if s < 0.001:
            aa.axis = Vector3(1.0, 0.0, 0.0)
        else:
            aa.axis.x = quaternion.x / s
            aa.axis.y = quaternion.y / s
            aa.axis.z = quaternion.z / s

        return aa

    @classmethod
    def from_null(
            cls
    ) -> AxisAngle:
        """
        Returns a AxisAngle that contains no rotation.

        Returns:
            A new AxisAngle instance.
        """
        return AxisAngle(Vector3.from_null(), 0.0)


class SurfaceMaterial(IntEnum):
    """
    The surface materials for walkmeshes found in both games.
    """
    # as according to 'surfacemat.2da'
    UNDEFINED = 0
    DIRT = 1
    OBSCURING = 2
    GRASS = 3
    STONE = 4
    WOOD = 5
    WATER = 6
    NON_WALK = 7
    TRANSPARENT = 8
    CARPET = 9
    METAL = 10
    PUDDLES = 11
    SWAMP = 12
    MUD = 13
    LEAVES = 14
    LAVA = 15
    BOTTOMLESS_PIT = 16
    DEEP_WATER = 17
    DOOR = 18
    NON_WALK_GRASS = 19
    SURFACE_MATERIAL_20 = 20
    SURFACE_MATERIAL_21 = 21
    SURFACE_MATERIAL_22 = 22
    SURFACE_MATERIAL_23 = 23
    SURFACE_MATERIAL_24 = 24
    SURFACE_MATERIAL_25 = 25
    SURFACE_MATERIAL_26 = 26
    SURFACE_MATERIAL_27 = 27
    SURFACE_MATERIAL_28 = 28
    SURFACE_MATERIAL_29 = 29
    TRIGGER = 30

    def walkable(
            self
    ) -> bool:
        return self in [SurfaceMaterial.DIRT, SurfaceMaterial.GRASS, SurfaceMaterial.STONE, SurfaceMaterial.WOOD,
                        SurfaceMaterial.WATER, SurfaceMaterial.CARPET, SurfaceMaterial.METAL, SurfaceMaterial.PUDDLES,
                        SurfaceMaterial.SWAMP, SurfaceMaterial.MUD, SurfaceMaterial.LEAVES, SurfaceMaterial.DOOR,
                        SurfaceMaterial.TRIGGER]


class Face:
    """
    Represents a triangle in 3D space.

    Attributes:
        v1: First point of the triangle.
        v2: Second point of the triangle.
        v3: Third point of the triangle.
        material: Material of the triangle, for usage in-game.
    """

    def __init__(
            self,
            v1: Vector3,
            v2: Vector3,
            v3: Vector3,
            material=SurfaceMaterial.UNDEFINED
    ):
        self.v1: Vector3 = v1
        self.v2: Vector3 = v2
        self.v3: Vector3 = v3
        self.material: SurfaceMaterial = material

    def normal(
            self
    ) -> Vector3:
        """
        Returns the normal for the face.

        Returns:
            A new Vector3 instance representing the face normal.
        """
        u = self.v2 - self.v1
        v = self.v3 - self.v2

        normal = Vector3.from_null()
        normal.x = (u.y * v.z) - (u.z * v.y)
        normal.y = (u.z * v.x) - (u.x * v.z)
        normal.z = (u.x * v.y) - (u.y * v.x)
        normal.normalize()

        return normal

    def area(
            self
    ) -> float:
        a = self.v1.distance(self.v2)
        b = self.v1.distance(self.v3)
        c = self.v2.distance(self.v3)
        return 0.25 * math.sqrt((a + b + c) * (-a + b + c) * (a - b + c) * (a + b - c))

    def planar_distance(
            self
    ) -> float:
        return -1.0 * (self.normal().dot(self.v1))

    def centre(
            self
    ) -> Vector3:
        return (self.v1 + self.v2 + self.v3) / 3

    def average(
            self
    ) -> Vector3:
        """
        Returns the average point of the face.

        Returns:
            A vector representing the average point of the face.
        """
        return (self.v1 + self.v2 + self.v3) / 3

    def determine_z(
            self,
            x: float,
            y: float
    ):
        """
        Returns the Z-component determined from the given X and Y components. This method does not check if the point
        exists within the face, that must be done seperatley with inside().

        Returns:
            The Z-component.
        """
        dx1 = x - self.v1.x
        dy1 = y - self.v1.y
        dx2 = self.v2.x - self.v1.x
        dy2 = self.v2.y - self.v1.y
        dx3 = self.v3.x - self.v1.x
        dy3 = self.v3.z - self.v1.y
        scale = dx3 * dy2 - dx2 * dy3
        nx = (dx1 * dy2 - dy1 * dx2) / scale
        ny = (dy1 * dx3 - dx1 * dy3) / scale
        return self.v1.z + ny * (self.v2.z - self.v1.z) + nx * (self.v3.z - self.v1.z)


class Polygon2:
    def __init__(
            self,
            points: List[Vector2] = None
    ):
        self.points: List[Vector2] = [] if points is None else points

    def __iter__(
            self
    ):
        for point in self.points:
            yield point

    def __len__(
            self
    ):
        return len(self.points)

    def __repr__(
            self
    ):
        return "Polygon3({})".format(self.points)

    def __getitem__(
            self,
            item: int
    ):
        if isinstance(item, int):
            return self.points[item]
        elif isinstance(item, slice):
            return self.points[item.start:item.stop:item.step]
        else:
            return NotImplemented

    def __setitem__(
            self,
            key: int,
            value: Vector2
    ):
        if isinstance(key, int) and isinstance(value, Vector2):
            self.points[key] = value
        else:
            return NotImplemented

    @classmethod
    def from_polygon3(
            cls,
            poly3: Polygon3
    ) -> Polygon2:
        """
        Converts a Polygon3 object into a Polygon2 object. The Z-axis is removed.

        Args:
            poly3: The polygon3 object to flatten.

        Returns:
            A Polygon2 object.
        """
        poly2 = Polygon2()
        for point in poly3:
            poly2.points.append(Vector2(point.x, point.y))
        return poly2

    def inside(
            self,
            point: Vector2,
            include_edges: bool = True
    ) -> bool:
        """
        Returns if the specified point is inside the polygon.

        Args:
            point: The point.
            include_edges: If True, a points on edges are considered inside the polygon.

        Returns:
            True if the point is inside the polygon.
        """
        # Code was adapted from this stackoverflow post:
        # https://stackoverflow.com/a/42306732

        n = len(self.points)
        inside = False

        p1 = self.points[0]
        for i in range(1, n + 1):
            p2 = self.points[i % n]
            if p1.y == p2.y:
                if point.y == p2.y:
                    if min(p1.x, p2.x) <= point.x <= max(p1.x, p2.x):
                        inside = include_edges
                        break
                    elif point.x < min(p1.x, p2.x):
                        inside = not inside
            else:
                if min(p1.y, p2.y) <= point.y <= max(p1.y, p2.y):
                    xinters = (point.y - p1.y) * (p2.x - p1.x) / float(p2.y - p1.y) + p1.x

                    if point.x == xinters:
                        inside = include_edges
                        break

                    if point.x < xinters:
                        inside = not inside
            p1 = p2
        return inside

    def area(
            self
    ) -> float:
        # Code was adapted from this stackoverflow post:
        # https://stackoverflow.com/a/34327761

        n = len(self.points)
        area = 0.0
        for i in range(0, n - 1):
            area += -self.points[i].y * self.points[i + 1].x + self.points[i].x * self.points[i + 1].y
        area += -self.points[n - 1].y * self.points[0].x + self.points[n - 1].x * self.points[0].y
        area = 0.5 * math.fabs(area)
        return area

    def append(
            self,
            point: Vector2
    ) -> None:
        self.points.append(point)

    def extend(
            self,
            points: List[Vector2]
    ) -> None:
        self.points.extend(points)

    def remove(
            self,
            point: Vector2
    ) -> None:
        self.points.remove(point)

    def index(
            self,
            point: Vector2
    ) -> int:
        return self.points.index(point)


class Polygon3:
    def __init__(
            self,
            points: List[Vector3] = None
    ):
        self.points: List[Vector3] = [] if points is None else points

    def __iter__(
            self
    ):
        for point in self.points:
            yield point

    def __len__(
            self
    ):
        return len(self.points)

    def __repr__(
            self
    ):
        return "Polygon3({})".format(self.points)

    def __getitem__(
            self,
            item: int
    ):
        if isinstance(item, int):
            return self.points[item]
        elif isinstance(item, slice):
            return self.points[item.start:item.stop:item.step]
        else:
            return NotImplemented

    def __setitem__(
            self,
            key: int,
            value: Vector3
    ):
        if isinstance(key, int) and isinstance(value, Vector3):
            self.points[key] = value
        else:
            return NotImplemented

    @classmethod
    def from_polygon2(
            cls,
            poly2: Polygon2
    ) -> Polygon3:
        """
        Converts a Polygon3 object into a Polygon2 object. Points have their Z-axis set to 0.

        Args:
            poly2: The polygon3 object to copy.

        Returns:
            A Polygon2 object.
        """
        poly3 = Polygon3()
        for point in poly3:
            poly3.points.append(Vector3(point.x, point.y, point.z))
        return poly3

    def append(
            self,
            point: Vector3
    ) -> None:
        self.points.append(point)

    def extend(
            self,
            points: List[Vector3]
    ) -> None:
        self.points.extend(points)

    def remove(
            self,
            point: Vector3
    ) -> None:
        self.points.remove(point)

    def index(
            self,
            point: Vector3
    ) -> int:
        return self.points.index(point)



from __future__ import annotations

from typing import Optional

from pykotor.resource.formats.twoda.twoda_data import TwoDA
from pykotor.resource.type import TARGET_TYPES, SOURCE_TYPES, ResourceReader, ResourceWriter, autoclose


class TwoDABinaryReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._twoda: Optional[TwoDA] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> TwoDA:
        self._twoda = TwoDA()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if file_type != "2DA ":
            raise TypeError("The file type that was loaded is invalid.")

        if file_version != "V2.b":
            raise TypeError("The 2DA version that was loaded is not supported.")

        self._reader.read_uint8()  # \n

        columns = []
        while self._reader.peek() != b'\0':
            column_header = self._reader.read_terminated_string("\t")
            self._twoda.add_column(column_header)
            columns.append(column_header)

        self._reader.read_uint8()  # \0

        row_count = self._reader.read_uint32()
        column_count = self._twoda.get_width()
        cell_count = row_count * column_count
        for i in range(row_count):
            row_header = self._reader.read_terminated_string("\t")
            row_label = row_header
            self._twoda.add_row(row_label)

        cell_offsets = [0] * cell_count
        for i in range(cell_count):
            cell_offsets[i] = self._reader.read_uint16()

        cell_data_size = self._reader.read_uint16()
        cell_data_offset = self._reader.position()

        for i in range(cell_count):
            column_id = i % column_count
            row_id = i // column_count
            column_header = columns[column_id]
            self._reader.seek(cell_data_offset + cell_offsets[i])
            cell_value = self._reader.read_terminated_string("\0")
            self._twoda.set_cell(row_id, column_header, cell_value)

        return self._twoda


class TwoDABinaryWriter(ResourceWriter):
    def __init__(
            self,
            twoda: TwoDA,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._twoda: TwoDA = twoda

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        headers = self._twoda.get_headers()

        self._writer.write_string("2DA ")
        self._writer.write_string("V2.b")

        self._writer.write_string("\n")
        for header in headers:
            self._writer.write_string(header + "\t")
        self._writer.write_string("\0")

        self._writer.write_uint32(self._twoda.get_height())
        for row_label in self._twoda.get_labels():
            self._writer.write_string(str(row_label) + "\t")

        values = []
        value_offsets = []
        cell_offsets = []
        data_size = 0

        for row in self._twoda:
            for header in self._twoda.get_headers():
                value = row.get_string(header) + "\0"
                if value not in values:
                    value_offset = 0 if len(value_offsets) == 0 else len(values[-1]) + value_offsets[-1]
                    values.append(value)
                    value_offsets.append(value_offset)
                    data_size += len(value)
                cell_offset = value_offsets[values.index(value)]
                cell_offsets.append(cell_offset)

        for cell_offset in cell_offsets:
            self._writer.write_uint16(cell_offset)
        self._writer.write_uint16(data_size)

        for value in values:
            self._writer.write_string(value)



from __future__ import annotations

import struct
from typing import Optional

from pykotor.common.geometry import Vector3, SurfaceMaterial
from pykotor.resource.formats.bwm import BWM, BWMFace, BWMType
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter, autoclose


class BWMBinaryReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._wok: Optional[BWM] = None
        self.position: Vector3 = Vector3.from_null()
        self.relative_hook1: Vector3 = Vector3.from_null()
        self.relative_hook2: Vector3 = Vector3.from_null()
        self.absolute_hook1: Vector3 = Vector3.from_null()
        self.absolute_hook2: Vector3 = Vector3.from_null()

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> BWM:
        self._wok = BWM()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if file_type != "BWM ":
            raise ValueError("Not a valid binary BWM file.")

        if file_version != "V1.0":
            raise ValueError("The BWM version of the file is unsupported.")

        self._wok.walkmesh_type = BWMType(self._reader.read_uint32())
        self._wok.relative_hook1 = self._reader.read_vector3()
        self._wok.relative_hook2 = self._reader.read_vector3()
        self._wok.absolute_hook1 = self._reader.read_vector3()
        self._wok.absolute_hook2 = self._reader.read_vector3()
        self._wok.position = self._reader.read_vector3()

        vertices_count = self._reader.read_uint32()
        vertices_offset = self._reader.read_uint32()
        face_count = self._reader.read_uint32()
        indices_offset = self._reader.read_uint32()
        materials_offset = self._reader.read_uint32()
        normals_offset = self._reader.read_uint32()
        planar_distances_offset = self._reader.read_uint32()

        aabb_count = self._reader.read_uint32()
        aabb_offset = self._reader.read_uint32()
        self._reader.skip(4)
        adjacencies_count = self._reader.read_uint32()
        adjacencies_offset = self._reader.read_uint32()
        edges_count = self._reader.read_uint32()
        edges_offset = self._reader.read_uint32()
        perimeters_count = self._reader.read_uint32()
        perimeters_offset = self._reader.read_uint32()

        vertices = []
        self._reader.seek(vertices_offset)
        for i in range(vertices_count):
            vertices.append(self._reader.read_vector3())

        faces = []
        self._reader.seek(indices_offset)
        for i in range(face_count):
            i1, i2, i3 = self._reader.read_uint32(), self._reader.read_uint32(), self._reader.read_uint32()
            v1, v2, v3 = vertices[i1], vertices[i2], vertices[i3]
            faces.append(BWMFace(v1, v2, v3))

        walkable_count = 0
        self._reader.seek(materials_offset)
        for face in faces:
            material_id = self._reader.read_uint32()
            face.material = SurfaceMaterial(material_id)
            if face.material.walkable():
                walkable_count += 1

        self._reader.seek(edges_offset)
        x = []
        for i in range(edges_count):
            edge_index = self._reader.read_uint32()
            x.append(edge_index)
            transition = self._reader.read_uint32()

            if transition != 0xFFFFFFFF:
                face_index = edge_index // 3
                trans_index = edge_index % 3
                if trans_index == 0:
                    faces[face_index].trans1 = transition
                if trans_index == 1:
                    faces[face_index].trans2 = transition
                if trans_index == 2:
                    faces[face_index].trans3 = transition

        self._wok.faces = faces

        return self._wok


class BWMBinaryWriter(ResourceWriter):
    HEADER_SIZE = 136

    def __init__(
            self,
            wok: BWM,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._wok: BWM = wok

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        vertices = self._wok.vertices()

        walkable = [face for face in self._wok.faces if face.material.walkable()]
        unwalkable = [face for face in self._wok.faces if not face.material.walkable()]
        faces = walkable + unwalkable
        aabbs = self._wok.aabbs()

        vertex_offset = 136
        vertex_data = bytearray()
        for vertex in vertices:
            vertex_data += struct.pack("fff", vertex.x, vertex.y, vertex.z)

        indices_offset = vertex_offset + len(vertex_data)
        indices_data = bytearray()
        for face in faces:
            i1, i2, i3 = vertices.index(face.v1), vertices.index(face.v2), vertices.index(face.v3)
            indices_data += struct.pack("III", i1, i2, i3)

        material_offset = indices_offset + len(indices_data)
        material_data = bytearray()
        for face in faces:
            material_data += struct.pack("I", face.material.value)

        normal_offset = material_offset + len(material_data)
        normal_data = bytearray()
        for face in faces:
            normal = face.normal()
            normal_data += struct.pack("fff", normal.x, normal.y, normal.z)

        coefficient_offset = normal_offset + len(normal_data)
        coeffeicent_data = bytearray()
        for face in faces:
            coeffeicent_data += struct.pack("f", face.planar_distance())

        aabb_offset = coefficient_offset + len(coeffeicent_data)
        aabb_data = bytearray()
        for aabb in aabbs:
            aabb_data += struct.pack("fff", aabb.bb_min.x, aabb.bb_min.y, aabb.bb_min.z)
            aabb_data += struct.pack("fff", aabb.bb_max.x, aabb.bb_max.y, aabb.bb_max.z)
            aabb_data += struct.pack("I", faces.index(aabb.face) if aabb.face is not None else 0xFFFFFFFF)
            aabb_data += struct.pack("I", 4)
            aabb_data += struct.pack("I", aabb.sigplane.value)
            aabb_data += struct.pack("I", aabbs.index(aabb.left) + 1 if aabb.left is not None else 0xFFFFFFFF)
            aabb_data += struct.pack("I", aabbs.index(aabb.right) + 1 if aabb.right is not None else 0xFFFFFFFF)

        adjacency_offset = aabb_offset + len(aabb_data)
        adjacency_data = bytearray()
        for face in walkable:
            adjancencies = self._wok.adjacencies(face)
            indexes = []
            for adjacency in adjancencies:
                indexes.append(faces.index(adjacency.face) * 3 + adjacency.edge if adjacency is not None else -1)
            adjacency_data += struct.pack("iii", *indexes)

        edges = self._wok.edges()
        edge_data = bytearray()
        edge_offset = adjacency_offset + len(adjacency_data)
        for edge in edges:
            edge_index = faces.index(edge.face) * 3 + edge.index
            edge_data += struct.pack("ii", edge_index, edge.transition)

        perimeters = [edges.index(edge) + 1 for edge in edges if edge.final]
        perimeter_data = bytearray()
        perimeter_offset = edge_offset + len(edge_data)
        for perimeter in perimeters:
            perimeter_data += struct.pack("I", perimeter)

        self._writer.write_string("BWM V1.0")
        self._writer.write_uint32(self._wok.walkmesh_type.value)
        self._writer.write_vector3(self._wok.relative_hook1)
        self._writer.write_vector3(self._wok.relative_hook2)
        self._writer.write_vector3(self._wok.absolute_hook1)
        self._writer.write_vector3(self._wok.absolute_hook2)
        self._writer.write_vector3(self._wok.position)

        self._writer.write_uint32(len(vertices))
        self._writer.write_uint32(vertex_offset)
        self._writer.write_uint32(len(faces))
        self._writer.write_uint32(indices_offset)
        self._writer.write_uint32(material_offset)
        self._writer.write_uint32(normal_offset)
        self._writer.write_uint32(coefficient_offset)
        self._writer.write_uint32(len(aabbs))
        self._writer.write_uint32(aabb_offset)
        self._writer.write_uint32(0)
        self._writer.write_uint32(len(self._wok.walkable_faces()))
        self._writer.write_uint32(adjacency_offset)
        self._writer.write_uint32(len(edges))
        self._writer.write_uint32(edge_offset)
        self._writer.write_uint32(len(perimeters))
        self._writer.write_uint32(perimeter_offset)

        self._writer.write_bytes(vertex_data)
        self._writer.write_bytes(indices_data)
        self._writer.write_bytes(material_data)
        self._writer.write_bytes(normal_data)
        self._writer.write_bytes(coeffeicent_data)
        self._writer.write_bytes(aabb_data)
        self._writer.write_bytes(adjacency_data)
        self._writer.write_bytes(edge_data)
        self._writer.write_bytes(perimeter_data)



import os.path
from configparser import ConfigParser
from enum import IntEnum
from typing import List, Dict, Optional

from pykotor.extract.capsule import Capsule

from pykotor.resource.formats.gff.gff_auto import bytes_gff

from pykotor.resource.formats.erf import read_erf, write_erf, ERF
from pykotor.common.stream import BinaryReader, BinaryWriter

from pykotor.extract.file import ResourceIdentifier
from pykotor.extract.installation import Installation, SearchLocation
from pykotor.resource.formats.gff import read_gff, write_gff
from pykotor.resource.formats.rim import read_rim, write_rim, RIM
from pykotor.resource.formats.ssf import read_ssf, write_ssf
from pykotor.resource.formats.tlk import TLK, read_tlk, write_tlk
from pykotor.resource.formats.twoda import read_2da, write_2da
from pykotor.tslpatcher.logger import PatchLogger
from pykotor.tslpatcher.mods.gff import ModificationsGFF
from pykotor.tslpatcher.memory import PatcherMemory
from pykotor.tslpatcher.mods.install import InstallFolder
from pykotor.tslpatcher.mods.ssf import ModificationsSSF
from pykotor.tslpatcher.mods.tlk import ModificationsTLK
from pykotor.tslpatcher.mods.twoda import Modifications2DA


class LogLevel(IntEnum):
    # Docstrings taken from ChangeEdit docs

    Nothing = 0
    """No feedback at all. The text from "info.rtf" will continue to be displayed during installation"""

    General = 1
    """Only general progress information will be displayed. Not recommended."""

    Errors = 2
    """General progress information is displayed, along with any serious errors encountered."""

    Warnings = 3
    """General progress information, serious errors and warnings are displayed. This is
    recommended for the release version of your mod."""

    Full = 4
    """Full feedback. On top of what is displayed at level 3, it also shows verbose progress
    information that may be useful for a Modder to see what is happening. Intended for
    Debugging."""


class PatcherConfig:
    def __init__(self):
        self.window_title: str = ""
        self.confirm_message: str = ""
        self.game_number: Optional[int] = None

        self.required_file: Optional[str] = None
        self.required_message: str = ""

        self.install_list: List[InstallFolder] = []
        self.patches_2da: List[Modifications2DA] = []
        self.patches_gff: List[ModificationsGFF] = []
        self.patches_ssf: List[ModificationsSSF] = []
        self.patches_tlk: ModificationsTLK = ModificationsTLK()

    def load(self, ini_text: str, append: TLK) -> None:
        from pykotor.tslpatcher.reader import ConfigReader

        ini = ConfigParser()
        ini.optionxform = str
        ini.read_string(ini_text)

        ConfigReader(ini, append).load(self)

    def patch_count(self) -> int:
        return len(self.patches_2da) + len(self.patches_gff) + len(self.patches_ssf) + 1 + len(self.install_list)


class PatcherNamespace:
    def __init__(self):
        self.namespace_id: str = ""
        self.ini_filename: str = ""
        self.info_filename: str = ""
        self.data_folderpath: str = ""
        self.name: str = ""
        self.description: str = ""


class ModInstaller:
    def __init__(self, mod_path: str, game_path: str, ini_file: str, logger: PatchLogger = None):
        self.game_path: str = game_path
        self.mod_path: str = mod_path
        self.ini_file: str = ini_file
        self.output_path: str = game_path
        self.log: PatchLogger = PatchLogger() if logger is None else logger

        self._config: Optional[PatcherConfig] = None

    def config(self) -> PatcherConfig:
        """
        Returns the PatcherConfig object associated with the mod installer. The object is created when the method is
        first called then cached for future calls.
        """

        if self._config is None:
            ini_text = BinaryReader.load_file(self.mod_path + "/" + self.ini_file).decode()
            append_tlk = read_tlk(self.mod_path + "/append.tlk") if os.path.exists(self.mod_path + "/append.tlk") else TLK()
            self._config = PatcherConfig()
            self._config.load(ini_text, append_tlk)

        return self._config

    def install(self) -> None:
        config = self.config()

        installation = Installation(self.game_path)
        memory = PatcherMemory()
        twodas = {}
        soundsets = {}
        templates = {}

        # Apply changes to dialog.tlk
        dialog_tlk = read_tlk(installation.path() + "dialog.tlk")
        config.patches_tlk.apply(dialog_tlk, memory)
        write_tlk(dialog_tlk, self.output_path + "/dialog.tlk")
        self.log.complete_patch()

        for folder in config.install_list:
            folder.apply(self.log, self.mod_path, self.output_path)
            self.log.complete_patch()

        # Apply changes to 2DA files
        for patch in config.patches_2da:
            resname, restype = ResourceIdentifier.from_path(patch.filename)
            search = installation.resource(resname, restype, [SearchLocation.OVERRIDE, SearchLocation.CUSTOM_FOLDERS], folders=[self.mod_path])
            twoda = twodas[patch.filename] = read_2da(search.data)

            self.log.add_note("Patching {}".format(patch.filename))
            patch.apply(twoda, memory)
            write_2da(twoda, "{}/override/{}".format(self.output_path, patch.filename))

            self.log.complete_patch()

        # Apply changes to SSF files
        for patch in config.patches_ssf:
            resname, restype = ResourceIdentifier.from_path(patch.filename)
            search = installation.resource(resname, restype, [SearchLocation.OVERRIDE, SearchLocation.CUSTOM_FOLDERS], folders=[self.mod_path])
            soundset = soundsets[patch.filename] = read_ssf(search.data)

            self.log.add_note("Patching {}".format(patch.filename))
            patch.apply(soundset, memory)
            write_ssf(soundset, "{}/override/{}".format(self.output_path, patch.filename))

            self.log.complete_patch()

        # Apply changes to GFF files
        for patch in config.patches_gff:
            resname, restype = ResourceIdentifier.from_path(patch.filename)

            scan_capsules = []
            if patch.destination.endswith(".rim") or patch.destination.endswith(".erf") or patch.destination.endswith(".mod"):
                scan_capsules.append(Capsule(self.output_path + "/" + patch.destination))

            search = installation.resource(
                resname,
                restype,
                [SearchLocation.OVERRIDE, SearchLocation.CUSTOM_FOLDERS, SearchLocation.CUSTOM_MODULES],
                folders=[self.mod_path],
                capsules=scan_capsules
            )
            template = templates[patch.filename] = read_gff(search.data)

            self.log.add_note("Patching {}".format(patch.filename))
            patch.apply(template, memory, self.log)
            self.write("{}/{}".format(self.output_path, patch.destination), patch.filename, bytes_gff(template))

            self.log.complete_patch()

    def write(self, filepath: str, filename: str, data: bytes) -> None:
        resname, restype = ResourceIdentifier.from_path(filename)
        if filepath.endswith(".rim"):
            rim = read_rim(BinaryReader.load_file(filepath)) if os.path.exists(filepath) else RIM()
            rim.set(resname, restype, data)
            write_rim(rim, filepath)
        elif filepath.endswith(".mod") or filepath.endswith(".erf"):
            erf = read_erf(BinaryReader.load_file(filepath)) if os.path.exists(filepath) else ERF()
            erf.set(resname, restype, data)
            write_erf(erf, filepath)
        else:
            BinaryWriter.dump(filepath, data)



from __future__ import annotations

import os.path
from contextlib import suppress
from copy import copy
from typing import List, TypeVar, Generic, Optional, Any

from pykotor.resource.formats.bwm.bwm_auto import bytes_bwm
from pykotor.resource.formats.erf import write_erf, read_erf, ERFType
from pykotor.resource.formats.rim import read_rim, write_rim

from pykotor.resource.formats.vis.vis_auto import bytes_vis

from pykotor.resource.formats.tpc.tpc_auto import bytes_tpc

from pykotor.common.misc import CaseInsensitiveDict

from pykotor.common.stream import BinaryReader, BinaryWriter
from pykotor.extract.capsule import Capsule
from pykotor.extract.file import ResourceIdentifier
from pykotor.extract.installation import Installation, SearchLocation
from pykotor.resource.formats.bwm import read_bwm
from pykotor.resource.formats.gff import read_gff
from pykotor.resource.formats.lyt import LYT
from pykotor.resource.formats.lyt.lyt_auto import read_lyt, bytes_lyt
from pykotor.resource.formats.mdl import MDL
from pykotor.resource.formats.tpc import read_tpc, TPC
from pykotor.resource.formats.vis import read_vis, VIS
from pykotor.resource.generics.are import ARE, read_are, bytes_are
from pykotor.resource.generics.dlg import read_dlg, bytes_dlg
from pykotor.resource.generics.git import GIT, read_git, bytes_git
from pykotor.resource.generics.ifo import IFO, read_ifo, bytes_ifo
from pykotor.resource.generics.pth import read_pth, bytes_pth, PTH
from pykotor.resource.generics.utc import UTC, read_utc, bytes_utc
from pykotor.resource.generics.utd import UTD, read_utd, bytes_utd
from pykotor.resource.generics.ute import UTE, read_ute, bytes_ute
from pykotor.resource.generics.uti import UTI, read_uti, bytes_uti
from pykotor.resource.generics.utm import UTM, read_utm, bytes_utm
from pykotor.resource.generics.utp import UTP, read_utp, bytes_utp
from pykotor.resource.generics.uts import read_uts, bytes_uts, UTS
from pykotor.resource.generics.utt import UTT, read_utt, bytes_utt
from pykotor.resource.generics.utw import UTW, read_utw, bytes_utw
from pykotor.resource.type import ResourceType
from pykotor.tools.model import list_textures, list_lightmaps

T = TypeVar('T')
SEARCH_ORDER = [SearchLocation.OVERRIDE, SearchLocation.CUSTOM_MODULES, SearchLocation.CHITIN]


class Module:
    def __init__(
            self,
            root: str,
            installation: Installation,
            custom_capsule: Optional[Capsule] = None
    ):
        self._installation = installation
        self._root = root = root.lower()

        self._capsules = [custom_capsule] if custom_capsule is not None else []
        self._capsules.extend([Capsule(installation.module_path() + module) for module in installation.module_names() if root in module.lower()])

        for capsule in self._capsules:
            if capsule.exists("module", ResourceType.IFO):
                ifo = read_gff(capsule.resource("module", ResourceType.IFO))
                self._id = ifo.root.get_resref("Mod_Entry_Area").get().lower()
                break
        else:
            raise ValueError("Unable to locate module IFO file for '{}'.".format(root))

        self.resources: CaseInsensitiveDict[ModuleResource] = CaseInsensitiveDict()
        self.reload_resources()

    @staticmethod
    def get_root(
            filepath: str
    ) -> str:
        """
        Returns the root name for a module from the given filepath (or filename). For example "danm13_s.rim" would
        become "danm13".

        Args:
            filepath: The filename or filepath of one of the module encapsulated file.

        Returns:
            The string for the root name of a module.
        """
        root = os.path.basename(filepath).replace(".rim", "").replace(".erf", "").replace(".mod", "").lower()
        roota = root[:5]
        rootb = root[5:]
        if "_" in rootb:
            rootb = rootb[:rootb.index("_")]
        return roota + rootb

    def capsules(
            self
    ) -> List[Capsule]:
        """
        Returns a copy of the capsules used by the module.

        Returns:
            A list of linked capsules.
        """
        return copy(self._capsules)

    def reload_resources(
            self
    ):
        # Look in module files
        for capsule in self._capsules:
            for resource in capsule:
                resname = resource.resname()
                restype = resource.restype()
                self.add_locations(resname, restype, [capsule.path()])

        # Look for LYT/VIS
        for resource in self._installation.chitin_resources():
            if resource.resname() == self._id:
                self.add_locations(resource.resname(), resource.restype(), [resource.filepath()])
        for directory in self._installation.override_list():
            for resource in self._installation.override_resources(directory):
                if resource.resname() == self._id:
                    self.add_locations(resource.resname(), resource.restype(), [resource.filepath()])

        # Any resource linked in the GIT not present in the module files
        original = self.git().active()
        look_for = []
        for location in self.git().locations():
            self.git().activate(location)
            git = self.git().resource()
            [look_for.append(ResourceIdentifier(creature.resref.get(), ResourceType.UTC)) for creature in git.creatures]
            [look_for.append(ResourceIdentifier(placeable.resref.get(), ResourceType.UTP)) for placeable in
             git.placeables]
            [look_for.append(ResourceIdentifier(door.resref.get(), ResourceType.UTD)) for door in git.doors]
            [look_for.append(ResourceIdentifier(sound.resref.get(), ResourceType.UTS)) for sound in git.sounds]
            [look_for.append(ResourceIdentifier(waypoint.resref.get(), ResourceType.UTW)) for waypoint in git.waypoints]
            [look_for.append(ResourceIdentifier(encounter.resref.get(), ResourceType.UTE)) for encounter in
             git.encounters]
            [look_for.append(ResourceIdentifier(trigger.resref.get(), ResourceType.UTT)) for trigger in git.triggers]
            [look_for.append(ResourceIdentifier(store.resref.get(), ResourceType.UTM)) for store in git.stores]
        self.git().activate(original)

        # Models referenced in LYTs
        original = self.layout().active()
        for location in self.layout().locations():
            self.layout().activate(location)
            layout = self.layout().resource()
            for room in layout.rooms:
                look_for.append(ResourceIdentifier(room.model, ResourceType.MDL))
                look_for.append(ResourceIdentifier(room.model, ResourceType.MDX))
                look_for.append(ResourceIdentifier(room.model, ResourceType.WOK))
        self.layout().activate(original)

        search = self._installation.locations(look_for, [SearchLocation.OVERRIDE, SearchLocation.CHITIN])
        for identifier, locations in search.items():
            self.add_locations(identifier.resname, identifier.restype, [location.filepath for location in locations])

        # Also try get paths for textures in models
        look_for = []
        textures = set()
        for model in self.models():
            with suppress(Exception):
                data = model.data()
                for texture in list_textures(data):
                    textures.add(texture)
                for lightmap in list_lightmaps(data):
                    textures.add(lightmap)
        for texture in textures:
            look_for.append(ResourceIdentifier(texture, ResourceType.TPC))
            look_for.append(ResourceIdentifier(texture, ResourceType.TGA))
        search = self._installation.locations(look_for, [SearchLocation.OVERRIDE, SearchLocation.CHITIN,
                                                         SearchLocation.TEXTURES_TPA,
                                                         SearchLocation.TEXTURES_TPB, SearchLocation.TEXTURES_TPC])
        for identifier, locations in search.items():
            if len(locations) == 0:
                continue
            self.add_locations(identifier.resname, identifier.restype, [location.filepath for location in locations])

        for resource in self.resources.values():
            resource.activate()

    def add_locations(
            self,
            resname: str,
            restype: ResourceType,
            locations: List[str]
    ):
        # In order to store TGA resources in the same ModuleResource as their TPC counterpart, we use the .TPC extension
        # instead of the .TGA for the dictionary key.
        filename_ext = str(ResourceType.TPC if restype == ResourceType.TGA else restype)
        filename = "{}.{}".format(resname, filename_ext)
        if filename in self.resources:
            self.resources[filename].add_locations(locations)
        else:
            self.resources[filename] = ModuleResource(resname, restype, self._installation)
            self.resources[filename].add_locations(locations)

    def installation(
            self
    ) -> Installation:
        return self._installation

    def resource(
            self,
            resname: str,
            restype: ResourceType
    ) -> Optional[ModuleResource]:
        filename = resname + "." + restype.extension
        return self.resources[filename] if filename in self.resources else None

    def layout(
            self
    ) -> ModuleResource[LYT]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.LYT:
                return resource

    def vis(
            self
    ) -> ModuleResource[VIS]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.VIS:
                return resource

    def are(
            self
    ) -> ModuleResource[ARE]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.ARE:
                return resource

    def git(
            self
    ) -> ModuleResource[GIT]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.GIT:
                return resource

    def pth(
            self
    ) -> ModuleResource[PTH]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.PTH:
                return resource

    def info(
            self
    ) -> ModuleResource[IFO]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == "module" and resource.restype() == ResourceType.IFO:
                return resource

    def creature(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTC]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTC:
                return resource
        return None

    def creatures(
            self
    ) -> List[ModuleResource[UTC]]:
        creatures = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTC:
                creatures.append(resource)
        return creatures

    def placeable(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTP]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTP:
                return resource
        return None

    def placeables(
            self
    ) -> List[ModuleResource[UTP]]:
        placeables = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTP:
                placeables.append(resource)
        return placeables

    def door(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTD]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTD:
                return resource
        return None

    def doors(
            self
    ) -> List[ModuleResource[UTD]]:
        doors = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTD:
                doors.append(resource)
        return doors

    def item(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTI]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTI:
                return resource
        return None

    def items(
            self
    ) -> List[ModuleResource[UTI]]:
        doors = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTD:
                doors.append(resource)
        return doors

    def encounter(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTE]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTE:
                return resource
        return None

    def encounters(
            self
    ) -> List[ModuleResource[UTE]]:
        encounters = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTE:
                encounters.append(resource)
        return encounters

    def store(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTM]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTM:
                return resource
        return None

    def stores(
            self
    ) -> List[ModuleResource[UTM]]:
        stores = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTM:
                stores.append(resource)
        return stores

    def trigger(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTT]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTT:
                return resource
        return None

    def triggers(
            self
    ) -> List[ModuleResource[UTT]]:
        triggers = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTT:
                triggers.append(resource)
        return triggers

    def waypoint(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTW]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTW:
                return resource
        return None

    def waypoints(
            self
    ) -> List[ModuleResource[UTW]]:
        waypoints = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTW:
                waypoints.append(resource)
        return waypoints

    def model(
            self,
            resname: str
    ) -> Optional[ModuleResource[MDL]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.MDL:
                return resource
        return None

    def model_ext(
            self,
            resname: str
    ) -> Optional[ModuleResource]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.MDX:
                return resource
        return None

    def models(
            self
    ) -> List[ModuleResource[MDL]]:
        models = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.MDL:
                models.append(resource)
        return models

    def texture(
            self,
            resname: str
    ) -> Optional[ModuleResource[TPC]]:
        for resource in self.resources.values():
            if resname.lower() == resource.resname().lower() and resource.restype() in [ResourceType.TPC, ResourceType.TGA]:
                return resource
        return None

    def textures(
            self
    ) -> List[ModuleResource[MDL]]:
        textures = []
        for resource in self.resources.values():
            if resource.restype() in [ResourceType.TPC, ResourceType.TGA]:
                textures.append(resource)
        return textures

    def sound(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTS]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTS:
                return resource
        return None

    def sounds(
            self
    ) -> List[ModuleResource[UTS]]:
        sounds = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTS:
                sounds.append(resource)
        return sounds


class ModuleResource(Generic[T]):
    def __init__(
            self,
            resname: str,
            restype: ResourceType,
            installation: Installation
    ):
        self._resname: str = resname
        self._installation = installation
        self._restype: ResourceType = restype
        self._active: Optional[str] = None
        self._resource: Any = None
        self._locations: List[str] = []

    def resname(
            self
    ) -> str:
        """
        Returns the resource name.

        Returns:
            The resource name.
        """
        return self._resname

    def restype(
            self
    ) -> ResourceType:
        """
        Returns the type of resource stored.

        Returns:
            The resource type.
        """
        return self._restype

    def localized_name(
            self
    ) -> Optional[str]:
        res = self.resource()
        if res is None:
            return None
        elif isinstance(res, UTC):
            return self._installation.string(res.first_name) + " " + self._installation.string(res.last_name)
        elif isinstance(res, UTP):
            return self._installation.string(res.name)
        elif isinstance(res, UTD):
            return self._installation.string(res.name)
        elif isinstance(res, UTW):
            return self._installation.string(res.name)
        elif isinstance(res, UTT):
            return self._installation.string(res.name)
        elif isinstance(res, UTE):
            return self._installation.string(res.name)
        elif isinstance(res, UTM):
            return self._installation.string(res.name)
        elif isinstance(res, UTS):
            return self._installation.string(res.name)
        else:
            return None

    def data(
            self
    ) -> bytes:
        """
        Opens the file at the active location and returns the data.

        Raises:
            ValueError: If no file is active.

        Returns:
            The bytes data of the active file.
        """

        if self._active is None:
            raise ValueError("No file is currently active for resource '{}.{}'.".format(self.resname, self._restype.extension))
        elif self._active.endswith(".erf") or self._active.endswith(".mod") or self._active.endswith(".rim"):
            capsule = Capsule(self._active)
            return capsule.resource(self._resname, self._restype)
        elif self._active.endswith("bif"):
            return self._installation.resource(self._resname, self._restype, [SearchLocation.CHITIN]).data
        else:
            return BinaryReader.load_file(self._active)

    def resource(
            self
    ) -> Optional[T]:
        """
        Returns the cached resource object. If no object has been cached, then it will load the object.

        Returns:
            The resource object.
        """

        if self._resource is None:
            conversions = {
                ResourceType.UTC: (lambda data: read_utc(data)),
                ResourceType.UTP: (lambda data: read_utp(data)),
                ResourceType.UTD: (lambda data: read_utd(data)),
                ResourceType.UTI: (lambda data: read_uti(data)),
                ResourceType.UTM: (lambda data: read_utm(data)),
                ResourceType.UTE: (lambda data: read_ute(data)),
                ResourceType.UTT: (lambda data: read_utt(data)),
                ResourceType.UTW: (lambda data: read_utw(data)),
                ResourceType.UTS: (lambda data: read_uts(data)),
                ResourceType.DLG: (lambda data: read_dlg(data)),
                ResourceType.PTH: (lambda data: read_pth(data)),
                ResourceType.NCS: (lambda data: data),
                ResourceType.TPC: (lambda data: read_tpc(data)),
                ResourceType.TGA: (lambda data: read_tpc(data)),
                ResourceType.LYT: (lambda data: read_lyt(data)),
                ResourceType.VIS: (lambda data: read_vis(data)),
                ResourceType.IFO: (lambda data: read_ifo(data)),
                ResourceType.ARE: (lambda data: read_are(data)),
                ResourceType.GIT: (lambda data: read_git(data)),
                ResourceType.WOK: (lambda data: read_bwm(data))
            }

            if self._active is None:
                self._resource = None
            elif self._active.endswith(".erf") or self._active.endswith(".mod") or self._active.endswith(".rim"):
                data = Capsule(self._active).resource(self._resname, self._restype)
                self._resource = conversions[self._restype](data)
            elif self._active.endswith("bif"):
                data = self._installation.resource(self._resname, self._restype, [SearchLocation.CHITIN]).data
                self._resource = conversions[self._restype](data)
            else:
                data = BinaryReader.load_file(self._active)
                self._resource = conversions[self._restype](data)

        return self._resource

    def add_locations(
            self,
            filepaths: List[str]
    ) -> None:
        """
        Adds a list of filepaths to the list of locations stored for the resource. If a filepath already exists, it is
        ignored.

        Args:
            filepaths: A list of filepaths pointing to a location for the resource.
        """
        self._locations.extend([filepath for filepath in filepaths if filepath not in self._locations])
        if self._active is None and self._locations:
            self.activate(self._locations[0])

    def locations(
            self,
    ) -> List[str]:
        return self._locations

    def activate(
            self,
            filepath: str = None
    ) -> None:
        """
        Sets the active file to the specified path. Calling this method will reset the loaded resource.

        Raises:
            ValueError: If the filepath is not stored in the resource list of locations.

        Args:
            filepath: The new active file.
        """
        self._resource = None
        if filepath is None:
            self._active = self._locations[0] if len(self._locations) > 0 else None
        elif filepath in self._locations:
            self._active = filepath
        else:
            raise ValueError("The filepath '{}' is not being tracked as a location for the resource.".format(self._active))

    def unload(
            self
    ) -> None:
        """
        Clears the cached resource object from memory.
        """
        self._resource = None

    def reload(
            self
    ) -> None:
        """
        Reloads the resource object from the active location.
        """
        self._resource = None
        self.resource()

    def active(
            self
    ) -> Optional[str]:
        """
        Returns the filepath of the currently active file for the resource.

        Returns:
            Filepath to the active resource.
        """
        return self._active

    def save(
            self
    ) -> None:
        conversions = {
            ResourceType.UTC: (lambda res: bytes_utc(res)),
            ResourceType.UTP: (lambda res: bytes_utp(res)),
            ResourceType.UTD: (lambda res: bytes_utd(res)),
            ResourceType.UTI: (lambda res: bytes_uti(res)),
            ResourceType.UTM: (lambda res: bytes_utm(res)),
            ResourceType.UTE: (lambda res: bytes_ute(res)),
            ResourceType.UTT: (lambda res: bytes_utt(res)),
            ResourceType.UTW: (lambda res: bytes_utw(res)),
            ResourceType.UTS: (lambda res: bytes_uts(res)),
            ResourceType.DLG: (lambda res: bytes_dlg(res)),
            ResourceType.PTH: (lambda res: bytes_pth(res)),
            ResourceType.NCS: (lambda res: res),
            ResourceType.TPC: (lambda res: bytes_tpc(res)),
            ResourceType.TGA: (lambda res: bytes_tpc(res)),
            ResourceType.LYT: (lambda res: bytes_lyt(res)),
            ResourceType.VIS: (lambda res: bytes_vis(res)),
            ResourceType.IFO: (lambda res: bytes_ifo(res)),
            ResourceType.ARE: (lambda res: bytes_are(res)),
            ResourceType.GIT: (lambda res: bytes_git(res)),
            ResourceType.WOK: (lambda res: bytes_bwm(res))
        }

        if self._active is None:
            raise ValueError("No active file selected for resource '{}.{}'".format(self._resname, self._restype.extension))
        elif self._active.endswith(".erf") or self._active.endswith(".mod"):
            erf = read_erf(self._active)
            erf.erf_type = ERFType.ERF if self._active.endswith(".erf") else ERFType.MOD
            erf.set(self._resname, self._restype, conversions[self._restype](self.resource()))
            write_erf(erf, self._active)
        elif self._active.endswith(".rim"):
            rim = read_rim(self._active)
            rim.set(self._resname, self._restype, conversions[self._restype](self.resource()))
            write_rim(rim, self._active)
        elif self._active.endswith("bif"):
            raise ValueError("Cannot save file to BIF.")
        else:
            BinaryWriter.dump(self._active, conversions[self._restype](self.resource()))



from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.jrl import construct_jrl, dismantle_jrl

TEST_FILE = "../../files/test.jrl"


class TestJRL(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        jrl = construct_jrl(gff)
        self.validate_io(jrl)

        gff = dismantle_jrl(jrl)
        jrl = construct_jrl(gff)
        self.validate_io(jrl)

    def validate_io(self, jrl):
        quest = jrl.quests[0]
        self.assertEqual("Plot to be considered worthy to hear the Sand People history.", quest.comment)
        self.assertEqual(33089, quest.name.stringref)
        self.assertEqual(4, quest.planet_id)
        self.assertEqual(72, quest.plot_index)
        self.assertEqual(1, quest.priority)
        self.assertEqual("Tat20aa_worthy", quest.tag)

        entry1 = quest.entries[0]
        self.assertFalse(entry1.end)
        self.assertEqual(10, entry1.entry_id)
        self.assertEqual(33090, entry1.text.stringref)
        self.assertAlmostEqual(5.0, entry1.xp_percentage, 1)

        entry2 = quest.entries[1]
        self.assertTrue(entry2.end)
        self.assertEqual(20, entry2.entry_id)
        self.assertEqual(33091, entry2.text.stringref)
        self.assertAlmostEqual(6.0, entry2.xp_percentage, 1)



from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.mdl import MDL, MDLBinaryReader, MDLBinaryWriter, MDLAsciiReader, MDLAsciiWriter
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceType


def detect_mdl(
        source: SOURCE_TYPES,
        offset: int = 0
) -> ResourceType:
    """
    Returns what format the MDL data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the MDL data.
        offset: Offset into the source data.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.

    Returns:
        The format of the MDL data.
    """
    try:
        if isinstance(source, str):
            with BinaryReader.from_file(source, offset) as reader:
                first4 = reader.read_bytes(4)
                file_format = ResourceType.MDL if first4 == b'\x00\x00\x00\x00' else ResourceType.MDL_ASCII
        elif isinstance(source, bytes) or isinstance(source, bytearray):
            file_format = ResourceType.MDL if source[:4] == b'\x00\x00\x00\x00' else ResourceType.MDL_ASCII
        elif isinstance(source, BinaryReader):
            first4 = source.read_bytes(4)
            file_format = ResourceType.MDL if first4 == b'\x00\x00\x00\x00' else ResourceType.MDL_ASCII
            source.skip(-4)
        else:
            file_format = ResourceType.INVALID
    except (FileNotFoundError, PermissionError, IsADirectoryError) as e:
        raise e
    except IOError:
        file_format = ResourceType.INVALID

    return file_format


def read_mdl(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = 0,
        source_ext: SOURCE_TYPES = None,
        offset_ext: int = 0,
        size_ext: int = 0
) -> MDL:
    """
    Returns an MDL instance from the source. The file format (MDL or MDL_ASCII) is automatically determined before
    parsing the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: The number of bytes to read from the source.
        source_ext: Source of the MDX data, if available.
        offset_ext: Offset into the source_ext data.
        size_ext: The number of bytes to read from the MDX source.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or the format could not be determined.

    Returns:
        An MDL instance.
    """
    file_format = detect_mdl(source, offset)

    if file_format is ResourceType.INVALID:
        raise ValueError("Failed to determine the format of the MDL file.")

    if file_format == ResourceType.MDL:
        return MDLBinaryReader(source, offset, size, source_ext, offset_ext, size_ext).load()
    elif file_format == ResourceType.MDL_ASCII:
        return MDLAsciiReader(source, offset, size).load()


def write_mdl(
        mdl: MDL,
        target: TARGET_TYPES,
        file_format: ResourceType = ResourceType.MDL,
        target_ext: TARGET_TYPES = None
) -> None:
    """
    Writes the MDL data to the target location with the specified format (MDL or MDL_ASCII).

    Args:
        mdl: The MDL file being written.
        target: The location to write the data to.
        file_format: The file format.
        target_ext: The location to write the MDX data to (if file format is MDL).

    Raises:
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.MDL:
        MDLBinaryWriter(mdl, target, target_ext).write()
    elif file_format == ResourceType.MDL_ASCII:
        MDLAsciiWriter(mdl, target).write()
    else:
        raise ValueError("Unsupported format specified; use MDL or MDL_ASCII.")



import traceback
from typing import Optional, Tuple

from pykotor.extract.installation import Installation

from pykotor.resource.type import ResourceType

from pykotor.resource.generics.uti import read_uti

from pykotor.common.misc import EquipmentSlot

from pykotor.resource.formats.twoda import TwoDA
from pykotor.resource.generics.utc import UTC


def get_body_model(
        utc: UTC,
        installation: Installation,
        *,
        appearance: Optional[TwoDA] = None,
        baseitems: Optional[TwoDA] = None
) -> Tuple[str, Optional[str]]:
    """
    Returns the model and texture names for the given creature.

    The value for the texture may be None and the default texture provided by the model should be used instead.

    If no value is specified for the appearance or baseitem parameters then they will be loaded from the given
    installation.

    Args:
        utc: UTC object of the target creature.
        installation: The relevant installation.
        appearance: The appearance.2da loaded into a TwoDA object.
        baseitems: The baseitems.2da loaded into a TwoDA object.

    Returns:
        Returns a tuple containing the name of the model and the texture to apply to the model.
    """
    if appearance is None:
        installation.resource("appearance", ResourceType.TwoDA)
    if baseitems is None:
        installation.resource("baseitems", ResourceType.TwoDA)

    body_model = ""
    override_texture = None

    if appearance.get_row(utc.appearance_id).get_string("modeltype") == "B":
        body_model = appearance.get_row(utc.appearance_id).get_string("modela")

        if utc.alignment <= 25:
            override_texture = appearance.get_row(utc.appearance_id).get_string("texaevil") + "01"
        else:
            override_texture = appearance.get_row(utc.appearance_id).get_string("texa") + "01"

        if EquipmentSlot.ARMOR in utc.equipment:

            armor_resref = utc.equipment[EquipmentSlot.ARMOR].resref.get()
            armor_uti = read_uti(installation.resource(armor_resref, ResourceType.UTI).data)
            armor_variation = baseitems.get_row(armor_uti.base_item).get_string("bodyvar").lower()

            normal_tex_column = "tex" + armor_variation
            evil_tex_column = "tex" + armor_variation + "evil"
            print(utc.alignment, evil_tex_column)
            if utc.alignment <= 25 and evil_tex_column in appearance.get_headers():
                tex_column = evil_tex_column
            else:
                tex_column = normal_tex_column

            model_column = "model" + armor_variation
            body_model = appearance.get_row(utc.appearance_id).get_string(model_column)
            override_texture = appearance.get_row(utc.appearance_id).get_string(tex_column) + str(armor_uti.texture_variation).rjust(2, "0")

    if body_model == "":
        body_model = appearance.get_row(utc.appearance_id).get_string("race")

    return body_model, override_texture


def get_weapon_models(
        utc: UTC,
        installation: Installation,
        *,
        appearance: Optional[TwoDA] = None,
        baseitems: Optional[TwoDA] = None
) -> Tuple[Optional[str], Optional[str]]:
    """
    Returns a tuple containing the right-hand weapon model and the left-hand weapon model (in that order).

    If no weapon is equipped in a particular hand the value will return None.

    If no value is specified for the appearance or baseitem parameters then they will be loaded from the given
    installation.

    Args:
        utc: UTC object of the target creature.
        installation: The relevant installation.
        appearance: The appearance.2da loaded into a TwoDA object.
        baseitems: The baseitems.2da loaded into a TwoDA object.

    Returns:
        Returns a tuple containing right-hand and left-hand weapon model names.
    """
    if appearance is None:
        installation.resource("appearance", ResourceType.TwoDA)
    if baseitems is None:
        installation.resource("baseitems", ResourceType.TwoDA)

    rhand_model = None
    lhand_model = None

    rhand_resref = utc.equipment[EquipmentSlot.RIGHT_HAND].resref.get() if EquipmentSlot.RIGHT_HAND in utc.equipment else None
    lhand_resref = utc.equipment[EquipmentSlot.LEFT_HAND].resref.get() if EquipmentSlot.LEFT_HAND in utc.equipment else None

    if rhand_resref is not None:
        rhand_uti = read_uti(installation.resource(rhand_resref, ResourceType.UTI).data)
        default_model = baseitems.get_row(rhand_uti.base_item).get_string("defaultmodel")
        rhand_model = default_model.replace("001", str(rhand_uti.model_variation).rjust(3, "0"))

    if lhand_resref is not None:
        lhand_uti = read_uti(installation.resource(lhand_resref, ResourceType.UTI).data)
        default_model = baseitems.get_row(lhand_uti.base_item).get_string("defaultmodel")
        lhand_model = default_model.replace("001", str(lhand_uti.model_variation).rjust(3, "0"))

    return rhand_model, lhand_model


def get_head_model(
        utc: UTC,
        installation: Installation,
        *,
        appearance: Optional[TwoDA] = None,
        heads: Optional[TwoDA] = None
) -> Tuple[Optional[str], Optional[str]]:
    """
    Returns the model and texture names for the head used by a creature.

    The value for the texture may be None and the default texture provided by the model should be used instead.

    If no value is specified for the appearance or heads parameters then they will be loaded from the given
    installation.

    Args:
        utc: UTC object of the target creature.
        installation: The relevant installation.
        appearance: The appearance.2da loaded into a TwoDA object.
        heads: The heads.2da loaded into a TwoDA object.

    Returns:
        Returns a tuple containing the name of the model and the texture to apply to the model.
    """
    if appearance is None:
        installation.resource("appearance", ResourceType.TwoDA)
    if heads is None:
        installation.resource("heads", ResourceType.TwoDA)

    model = None
    texture = None

    head_id = appearance.get_row(utc.appearance_id).get_integer("normalhead")
    if head_id is not None:
        model = heads.get_row(head_id).get_string("head")
        if utc.alignment < 10:
            texture = heads.get_row(head_id).get_string("headtexvvve")
        elif utc.alignment < 20:
            texture = heads.get_row(head_id).get_string("headtexvve")
        elif utc.alignment < 30:
            texture = heads.get_row(head_id).get_string("headtexve")
        elif utc.alignment < 40:
            texture = heads.get_row(head_id).get_string("headtexe")

        if texture == "":
            texture = None

    return model, texture


def get_mask_model(
        utc: UTC,
        installation: Installation,
) -> Optional[str]:
    """
    Returns the model for the mask a creature is wearing.

    The value for the texture will return None if the creature does not have a mask equipped.

    If no value is specified for the appearance or heads parameters then they will be loaded from the given
    installation.

    Args:
        utc: UTC object of the target creature.
        installation: The relevant installation.

    Returns:
        Returns a name of the mask model.
    """
    model = None

    if EquipmentSlot.HEAD in utc.equipment:
        resref = utc.equipment[EquipmentSlot.HEAD].resref.get()
        uti = read_uti(installation.resource(resref, ResourceType.UTI).data)
        model = "I_Mask_" + str(uti.model_variation).rjust(3, "0")

    return model
